#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility class handling which community/technology partner files for em_om reader."""

# pylint: disable=no-member,duplicate-code

from typing import Tuple, Dict, List


class EmOmUseCaseSelector:  # pylint: disable=too-few-public-methods
    """Decision maker about what needs to be parsed given arbitrary input.

    Users might invoke this dataconverter with arbitrary input, no input, or
    too much input. The UseCaseSelector decides what to do in each case.
    """

    def __init__(self, file_paths: Tuple[str] = None):
        """Analyze with which parser(s) (if any) input can be handled."""
        self.mime_types: Dict[str, list] = {}
        self.eln_parser_type: str = "none"
        self.eln: List[str] = []
        self.dat_parser_type: str = "none"
        self.dat: List[str] = []
        self.is_valid = False

        self.analyze_mime_types(file_paths)
        self.identify_case()

    def analyze_mime_types(self, file_paths: Tuple[str] = None):
        """Accept/reject and organize input-files based on mime-type."""
        self.supported_mime_types = ["yaml", "yml", "h5oina", "mtex", "zip", "dream3d"]
        # "score", "qube", "paradis", "ang", "cpr", "crc", "damask via dream3d"
        for mime_type in self.supported_mime_types:
            self.mime_types[mime_type] = []
        for file_name in file_paths:
            index = file_name.lower().rfind(".")
            if index >= 0:
                suffix = file_name.lower()[index + 1::]
                add = (suffix in self.supported_mime_types) \
                    and (file_name not in self.mime_types[suffix])
                if add is True:
                    self.mime_types[suffix].append(file_name)
        print(self.mime_types)

    def identify_case(self):
        """Identify which sub-parsers to use if any based on input mime_types."""
        # eln
        self.is_valid = False
        for mime_type in ["yaml", "yml"]:
            self.eln += self.mime_types[mime_type]
        if len(self.eln) == 1:
            self.eln_parser_type = "generic"
            self.is_valid = True
        else:
            self.eln_parser_type = "none"

        # (heavy) data from community or technology partners
        if self.needs_orix_parser() is True:
            self.dat += self.mime_types["h5oina"]
            if len(self.dat) == 1:
                self.dat_parser_type = "orix"
        elif self.needs_mtex_parser() is True:
            self.dat += self.mime_types["mtex"]
            if len(self.dat) == 1:
                self.dat_parser_type = "mtex"
        elif self.needs_zip_parser() is True:
            self.dat += self.mime_types["zip"]
            if len(self.dat) == 1:
                self.dat_parser_type = "zip"
        elif self.needs_dreamthreed_parser() is True:
            self.dat += self.mime_types["dream3d"]
            if len(self.dat) == 1:
                self.dat_parser_type = "dream3d"
        # elif self.needs_kikuchipy_parser() is True:
        #     self.dat_parser_type = "kikuchipy"
        # elif self.needs_score_parser() is True:
        #     self.dat_parser_type = "score"
        # elif self.needs_qube_parser() is True:
        #     self.dat_parser_type = "qube"
        # elif self.needs_paradis_parser() is True:
        #     self.dat_parser_type = "paradis"
        # elif self.needs_brinckmann_parser() is True:
        #     self.dat_parser_type = "brinckmann"
        else:
            self.dat_parser_type = "none"

        # if self.eln_parser_type != "none":  # and (self.dat_parser_type != "none"):
        print("Input suggests to use the following sub-parsers:")
        print(f"ELN parser: {self.eln_parser_type}")
        print(self.eln)
        print(f"Data parser: {self.dat_parser_type}")
        print(self.dat)
        # self.is_valid = True
        print(f"Input suggests that parsing is valid: {self.is_valid}")

    def needs_orix_parser(self):
        """Check if input suggests to use orix parser or not."""
        if "h5oina" in self.mime_types:
            if len(self.mime_types["h5oina"]) == 1:
                return True
        return False

    def needs_mtex_parser(self):
        """Check if input suggests to use MTex parser or not."""
        if "mtex" in self.mime_types:
            if len(self.mime_types["mtex"]) == 1:
                return True
        return False

    def needs_zip_parser(self):
        """Check if input suggests to use zipfile parser for generic pattern or not."""
        if "mtex" in self.mime_types:
            if len(self.mime_types["zip"]) == 1:
                return True
        return False

    def needs_dreamthreed_parser(self):
        """Check if input suggests to use DREAM.3D parser or not."""
        if "dream3d" in self.mime_types:
            if len(self.mime_types["dream3d"]) == 1:
                return True
        return False

    def needs_kikuchipy_parser(self):
        """Check if input suggests to use kikuchipy parser or not."""
        return False

    def needs_score_parser(self):
        """Check if input suggests to use SCORE cellular automaton model parser or not."""
        return False

    def needs_qube_parser(self):
        """Check if input suggests to use Bruker QUBE parser or not."""
        return False

    def needs_paradis_parser(self):
        """Check if input suggests to use ParaDIS DDD parser or not."""
        return False

    def needs_brinckmann_parser(self):
        """Check if input suggests to use S. Brinckmann's EBSD parser or not."""
        return False

    #     """Check if input suggests to use Brinckmann parser or not."""
    #     if len(self.mime_types["cpr"]) == 1 and len(self.mime_types["crc"]) == 1:
    #         # HKL Channel 5 cpr/crc file should be a pair with the same file name
    #         if self.mime_types["cpr"][0].lower().replace(".cpr", "") \
    #             == self.mime_types["crc"][0].lower().replace(".crc", ""):
    #             # add files to self.dat
    #             return True
    #     if len(self.mime_types["ang"]) == 1:
    #         # consider in the future to use rather the kikuchipy or MTex parser
    #         # add files to self.dat
    #         return True
    #     return False



# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A generic class for generating duplicate outputs for ellipsometry"""
import random
import numpy as np
import ase
from pynxtools.dataconverter.helpers import extract_atom_types


class MockEllips():
    """ A generic class for generating duplicate outputs for ELLIPSOMETRY

        Contains:
        - mock_sample:
            Chooses random entry from sample_list, overwrites sample_name
            and extracts atom_types
        - mock_chemical_formula:
            Creates a list of chemical formulas consisting of two atom types
        - modify_spectral_range:
            Change spectral range (i.e. wavelength array) and step size.
        - mock_angles:
            Change value and number of incident angles
        - choose_data_type:
            Chooses random entry from data_types
        - mock_signals:
            Mock data if data_type is Psi/Delta or tan(Psi)/cos(Delta)
        - mock_mueller_matrix:
            Mock data if data_type is Mueller matrix
        - mock_template:
            Creates mock ellipsometry data (by calling the above routines)
    """

    def __init__(self, data_template) -> None:
        self.data = data_template["measured_data"]
        self.wavelength = data_template["data_collection/NAME_spectrum[wavelength_spectrum]"]
        self.atom_types = data_template["atom_types"]
        self.sample_list: list = []
        self.data_types = ["Psi/Delta", "tan(Psi)/cos(Delta)", "Mueller matrix"]
        self.angles: list = []
        self.number_of_signals = 0

    def mock_sample(self, data_template) -> None:
        """ Chooses random entry from sample_list, overwrites sample_name
            and extracts atom_types
        """
        self.mock_chemical_formula()
        name = random.choice(self.sample_list)
        self.atom_types = extract_atom_types(name)
        data_template["sample_name"] = name
        data_template["atom_types"] = self.atom_types
        data_template["layer_structure"] = f"{name} bulk"
        data_template["experiment_description"] = f"RC2 scan on {name} bulk"

    def choose_data_type(self, data_template) -> None:
        """ Chooses random entry from data_types
        """
        data_type = random.choice(self.data_types)
        data_template["data_type"] = data_type
        if data_type == "Mueller matrix":
            self.number_of_signals = 16
        elif data_type in ("Psi/Delta", "tan(Psi)/cos(Delta)"):
            self.number_of_signals = 2

    def mock_chemical_formula(self) -> None:
        """ Creates a list of chemical formulas consisting of two atom types """
        part_1 = ase.atom.chemical_symbols[1:]
        part_2 = list(range(2, 20, 1))

        for el1 in part_1:
            for na1 in part_2:
                for el2 in part_1:
                    for na2 in part_2:
                        chemical_formula = f"{el1}{na1}{el2}{na2}"
                        if el1 != el2:
                            self.sample_list.append(chemical_formula)

    def mock_angles(self, data_template) -> None:
        """ Change value and number of incident angles
        """
        angle_list = [40, 45, 50, 55, 60, 65, 70, 75, 80]
        for _ in range(random.randrange(1, 4)):
            angle = random.choice(angle_list)
            self.angles.append(angle)
            angle_list.remove(angle)
        self.angles.sort()
        data_template["angle_of_incidence"] = self.angles
        if self.number_of_signals == 2:
            self.mock_signals(data_template)
        elif self.number_of_signals == 16:
            self.mock_mueller_matrix(data_template)

    def mock_signals(self, data_template) -> None:
        """ Mock data if data_type is Psi/Delta or tan(Psi)/cos(Delta)
            considering the (new) number of incident angles
        """
        my_numpy_array = np.empty([
            len(self.angles),
            self.number_of_signals,
            len(self.wavelength)
        ])
        for index in range(0, len(self.angles)):
            noise = np.random.normal(0, 0.5, self.data[0, 0, :].size)
            my_numpy_array[index] = self.data[0] * random.uniform(0.5, 1.5) + noise
        self.data = my_numpy_array
        data_template["measured_data"] = my_numpy_array

    def mock_mueller_matrix(self, data_template) -> None:
        """ Mock data if data_type is Mueller matrix (i.e. 16 elements/signals)
            considering the (new) number of incident angles
        """
        my_numpy_array = np.empty([
            len(self.angles),
            self.number_of_signals,
            len(self.wavelength)
        ])
        for idx in range(0, len(self.angles)):
            noise = np.random.normal(0, 0.1, self .data[0, 0, :].size)
            for m_idx in range(1, self.number_of_signals):
                my_numpy_array[idx][m_idx] = self.data[0][0] * random.uniform(.5, 1.)
                my_numpy_array[idx][m_idx] += noise
            my_numpy_array[idx][0] = my_numpy_array[0][0] / my_numpy_array[0][0]
        data_template["measured_data"] = my_numpy_array

    def modify_spectral_range(self, data_template) -> None:
        """ Change spectral range (i.e. wavlength array) and step size,
            while length of the wavelength array remains the same.
        """
        temp = random.uniform(0.25, 23)
        data_template["data_collection/NAME_spectrum[wavelength_spectrum]"] = \
            temp * data_template["data_collection/NAME_spectrum[wavelength_spectrum]"]

    def mock_template(self, data_template) -> None:
        """ Creates a mock ellipsometry template """
        self.mock_sample(data_template)
        self.modify_spectral_range(data_template)
        self.choose_data_type(data_template)
        self.mock_angles(data_template)



import os
from pathlib import Path

from ..utils.types import PathLike


def get_source_root() -> Path:
    """Root directory of the source code for the documentation
    (git repository root by default)"""
    if _SOURCE_ROOT is None:
        set_source_root(Path(__file__).absolute().parent.parent.parent)
    return _SOURCE_ROOT


def set_source_root(root: PathLike) -> None:
    if not isinstance(root, Path):
        root = Path(root)
    global _SOURCE_ROOT
    _SOURCE_ROOT = root.absolute()


def get_nxdl_root() -> Path:
    """Root directory of the XSD and NXDL files
    (same as the source root by default)"""
    if _NXDL_ROOT is None:
        set_nxdl_root(get_source_root())
    return _NXDL_ROOT


def set_nxdl_root(root: PathLike) -> None:
    if not isinstance(root, Path):
        root = Path(root)
    assert (root / _XSD_FILE_NAME).exists()
    global _NXDL_ROOT
    _NXDL_ROOT = root.absolute()


def get_xsd_file() -> Path:
    """The XSD file that defines the NXDL"""
    return get_nxdl_root() / _XSD_FILE_NAME


def get_xsd_units_file() -> Path:
    """The XSD file that defines the units in the NXDL"""
    return get_nxdl_root() / _XSD_UNITS_FILE_NAME


def get_nxdl_version_file() -> Path:
    """The version of the NeXus standard and the NeXus Definition language"""
    return get_source_root() / _VERSION_FILE_NAME


def get_build_root() -> Path:
    """Root directory in which the sources for documentation
    building are generated."""
    if _BUILD_ROOT is None:
        set_build_root(get_source_root() / "build")
    return _BUILD_ROOT


def set_build_root(root: PathLike) -> None:
    if not isinstance(root, Path):
        root = Path(root)
    global _BUILD_ROOT
    _BUILD_ROOT = root.absolute()


def manual_source_root() -> Path:
    """Source directory of the NeXus User Manual"""
    return get_source_root() / "manual"


def manual_build_root() -> Path:
    """Build directory of the NeXus User Manual"""
    return get_build_root() / "manual"


def impatient_source_root() -> Path:
    """Source directory of the NeXus Impatient Guide"""
    return get_source_root() / "impatient-guide"


def impatient_build_root() -> Path:
    """Build directory of the NeXus Impatient Guide"""
    return get_build_root() / "impatient-guide"


def manual_source_sphinxroot() -> Path:
    """Sphinx source directory of the NeXus User Manual"""
    return manual_source_root() / "source"


def manual_build_sphinxroot() -> Path:
    """Sphinx source directory for building of the NeXus User Manual"""
    return manual_build_root() / "source"


def manual_build_staticroot() -> Path:
    """Static source directory for building of the NeXus User Manual"""
    return manual_build_sphinxroot() / "_static"


def nxclass_build_root(nxdl_file: Path) -> Path:
    """NeXus class documentation directory for building of the NeXus User Manual"""
    root = manual_build_sphinxroot() / "classes"
    root /= os.path.relpath(nxdl_file.parent, get_nxdl_root())
    return root


def get_rst_filename(nxdl_file: Path) -> Path:
    rst_file_name = nxclass_build_root(nxdl_file)
    rst_file_name /= nxdl_file.with_suffix("").with_suffix(".rst").name
    return rst_file_name


_XSD_FILE_NAME = "nxdl.xsd"
_XSD_UNITS_FILE_NAME = "nxdlTypes.xsd"
_VERSION_FILE_NAME = "NXDL_VERSION"
_NXDL_ROOT = None
_BUILD_ROOT = None
_SOURCE_ROOT = None



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility class to analyze which vendor/community files are passed to em reader."""

# pylint: disable=no-member,duplicate-code

from typing import Tuple, Dict, List


class ApmUseCaseSelector:  # pylint: disable=too-few-public-methods
    """Decision maker about what needs to be parsed given arbitrary input.

    Users might invoke this dataconverter with arbitrary input, no input, or
    too much input. The UseCaseSelector decide what to do in each case.
    """

    def __init__(self, file_paths: Tuple[str] = None):
        """Initialize the class.

        dataset injects numerical data and metadata from an analysis.
        eln injects additional metadata and eventually numerical data.
        """
        self.case: Dict[str, list] = {}
        self.eln: List[str] = []
        self.cfg: List[str] = []
        self.reconstruction: List[str] = []
        self.ranging: List[str] = []
        self.is_valid = False
        self.supported_mime_types = [
            "pos", "epos", "apt", "rrng", "rng", "txt", "yaml", "yml"]
        for mime_type in self.supported_mime_types:
            self.case[mime_type] = []

        self.sort_files_by_mime_type(file_paths)
        self.check_validity_of_file_combinations()

    def sort_files_by_mime_type(self, file_paths: Tuple[str] = None):
        """Sort all input-files based on their mimetype to prepare validity check."""
        for file_name in file_paths:
            index = file_name.lower().rfind(".")
            if index >= 0:
                suffix = file_name.lower()[index + 1::]
                if suffix in self.supported_mime_types:
                    if file_name not in self.case[suffix]:
                        self.case[suffix].append(file_name)

    def check_validity_of_file_combinations(self):
        """Check if this combination of types of files is supported."""
        recon_input = 0  # reconstruction relevant file e.g. POS, ePOS, APT
        range_input = 0  # ranging definition file, e.g. RNG, RRNG
        other_input = 0  # generic ELN or OASIS-specific configurations
        for mime_type, value in self.case.items():
            if mime_type in ["pos", "epos", "apt"]:
                recon_input += len(value)
            elif mime_type in ["rrng", "rng", "txt"]:
                range_input += len(value)
            elif mime_type in ["yaml", "yml"]:
                other_input += len(value)
            else:
                continue

        if (recon_input == 1) and (range_input == 1) and (1 <= other_input <= 2):
            self.is_valid = True
            self.reconstruction: List[str] = []
            self.ranging: List[str] = []
            for mime_type in ["pos", "epos", "apt"]:
                self.reconstruction += self.case[mime_type]
            for mime_type in ["rrng", "rng", "txt"]:
                self.ranging += self.case[mime_type]
            yml: List[str] = []
            for mime_type in ["yaml", "yml"]:
                yml += self.case[mime_type]
            for entry in yml:
                if entry.endswith(".oasis.specific.yaml") \
                        or entry.endswith(".oasis.specific.yml"):
                    self.cfg += [entry]
                else:
                    self.eln += [entry]



# pylint: disable=too-many-lines
"""Parse NeXus definition files
"""

import os
import xml.etree.ElementTree as ET
from functools import lru_cache
from glob import glob
import textwrap


class NxdlAttributeError(Exception):
    """An exception for throwing an error when an Nxdl attribute is not found."""


def get_app_defs_names():
    """Returns all the AppDef names without their extension: .nxdl.xml"""
    app_def_path_glob = f"{get_nexus_definitions_path()}{os.sep}applications{os.sep}*.nxdl*"
    contrib_def_path_glob = (f"{get_nexus_definitions_path()}{os.sep}"
                             f"contributed_definitions{os.sep}*.nxdl*")
    files = sorted(glob(app_def_path_glob)) + sorted(glob(contrib_def_path_glob))
    return [os.path.basename(file).split(".")[0] for file in files] + ["NXroot"]


@lru_cache(maxsize=None)
def get_xml_root(file_path):
    """Reducing I/O time by caching technique"""

    return ET.parse(file_path).getroot()


def get_nexus_definitions_path():
    """Check NEXUS_DEF_PATH variable.
If it is empty, this function is filling it"""
    try:  # either given by sys env
        return os.environ['NEXUS_DEF_PATH']
    except KeyError:  # or it should be available locally under the dir 'definitions'
        local_dir = os.path.abspath(os.path.dirname(__file__))
        return os.path.join(local_dir, f"..{os.sep}definitions")


def get_hdf_root(hdf_node):
    """Get the root HDF5 node"""
    node = hdf_node
    while node.name != '/':
        node = node.parent
    return node


def get_hdf_parent(hdf_info):
    """Get the parent of an hdf_node in an hdf_info"""
    if 'hdf_path' not in hdf_info:
        return hdf_info['hdf_node'].parent
    node = get_hdf_root(hdf_info['hdf_node']) if 'hdf_root' not in hdf_info \
        else hdf_info['hdf_root']
    for child_name in hdf_info['hdf_path'].split('/'):
        node = node[child_name]
    return node


def get_parent_path(hdf_name):
    """Get parent path"""
    return '/'.join(hdf_name.split('/')[:-1])


def get_hdf_info_parent(hdf_info):
    """Get the hdf_info for the parent of an hdf_node in an hdf_info"""
    if 'hdf_path' not in hdf_info:
        return {'hdf_node': hdf_info['hdf_node'].parent}
    node = get_hdf_root(hdf_info['hdf_node']) if 'hdf_root' not in hdf_info \
        else hdf_info['hdf_root']
    for child_name in hdf_info['hdf_path'].split('/')[1:-1]:
        node = node[child_name]
    return {'hdf_node': node, 'hdf_path': get_parent_path(hdf_info['hdf_path'])}


def get_nx_class(nxdl_elem):
    """Get the nexus class for a NXDL node"""
    if 'category' in nxdl_elem.attrib.keys():
        return None
    try:
        return nxdl_elem.attrib['type']
    except KeyError:
        return 'NX_CHAR'


def get_nx_namefit(hdf_name, name, name_any=False):
    """Checks if an HDF5 node name corresponds to a child of the NXDL element
uppercase letters in front can be replaced by arbitraty name, but
uppercase to lowercase match is preferred,
so such match is counted as a measure of the fit"""
    if name == hdf_name:
        return len(name) * 2
    # count leading capitals
    counting = 0
    while counting < len(name) and name[counting].upper() == name[counting]:
        counting += 1
    if name_any or counting == len(name) or \
       (counting > 0 and hdf_name.endswith(name[counting:])):  # if potential fit
        # count the matching chars
        fit = 0
        for i in range(min(counting, len(hdf_name))):
            if hdf_name[i].upper() == name[i]:
                fit += 1
            else:
                break
        if fit == min(counting, len(hdf_name)):  # accept only full fits as better fits
            return fit
        return 0
    return -1      # no fit


def get_nx_classes():
    """Read base classes from the NeXus definition folder.
Check each file in base_classes, applications, contributed_definitions.
If its category attribute is 'base', then it is added to the list. """
    base_classes = sorted(glob(os.path.join(get_nexus_definitions_path(),
                                            'base_classes', '*.nxdl.xml')))
    applications = sorted(glob(os.path.join(get_nexus_definitions_path(),
                                            'applications', '*.nxdl.xml')))
    contributed = sorted(glob(os.path.join(get_nexus_definitions_path(),
                                           'contributed_definitions', '*.nxdl.xml')))
    nx_clss = []
    for nexus_file in base_classes + applications + contributed:
        root = get_xml_root(nexus_file)
        if root.attrib['category'] == 'base':
            nx_clss.append(str(nexus_file[nexus_file.rindex(os.sep) + 1:])[:-9])
    nx_clss = sorted(nx_clss)
    return nx_clss


def get_nx_units():
    """Read unit kinds from the NeXus definition/nxdlTypes.xsd file"""
    filepath = f"{get_nexus_definitions_path()}{os.sep}nxdlTypes.xsd"
    root = get_xml_root(filepath)
    units_and_type_list = []
    for child in root:
        for i in child.attrib.values():
            units_and_type_list.append(i)
    flag = False
    for line in units_and_type_list:
        if line == 'anyUnitsAttr':
            flag = True
            nx_units = []
        elif 'NX' in line and flag is True:
            nx_units.append(line)
        elif line == 'primitiveType':
            flag = False
        else:
            pass
    return nx_units


def get_nx_attribute_type():
    """Read attribute types from the NeXus definition/nxdlTypes.xsd file"""
    filepath = get_nexus_definitions_path() + '/nxdlTypes.xsd'
    root = get_xml_root(filepath)
    units_and_type_list = []
    for child in root:
        for i in child.attrib.values():
            units_and_type_list.append(i)
    flag = False
    for line in units_and_type_list:
        if line == 'primitiveType':
            flag = True
            nx_types = []
        elif 'NX' in line and flag is True:
            nx_types.append(line)
        elif line == 'anyUnitsAttr':
            flag = False
        else:
            pass
    return nx_types


def get_node_name(node):
    '''Node - xml node. Returns html documentation name.
    Either as specified by the 'name' or taken from the type (nx_class).
    Note that if only class name is available, the NX prefix is removed and
    the string is converted to UPPER case.'''
    if 'name' in node.attrib.keys():
        name = node.attrib['name']
    else:
        name = node.attrib['type']
        if name.startswith('NX'):
            name = name[2:].upper()
    return name


def belongs_to(nxdl_elem, child, name, class_type=None, hdf_name=None):
    """Checks if an HDF5 node name corresponds to a child of the NXDL element
uppercase letters in front can be replaced by arbitraty name, but
uppercase to lowercase match is preferred"""
    if class_type and get_nx_class(child) != class_type:
        return False
    act_htmlname = get_node_name(child)
    chk_name = hdf_name or name
    if act_htmlname == chk_name:
        return True
    if not hdf_name:  # search for name fits is only allowed for hdf_nodes
        return False
    try:  # check if nameType allows different name
        name_any = bool(child.attrib['nameType'] == "any")
    except KeyError:
        name_any = False
    params = [act_htmlname, chk_name, name_any, nxdl_elem, child, name]
    return belongs_to_capital(params)


def belongs_to_capital(params):
    """Checking continues for Upper case"""
    (act_htmlname, chk_name, name_any, nxdl_elem, child, name) = params
    # or starts with capital and no reserved words used
    if (name_any or 'A' <= act_htmlname[0] <= 'Z') and \
            name != 'doc' and name != 'enumeration':
        fit = get_nx_namefit(chk_name, act_htmlname, name_any)  # check if name fits
        if fit < 0:
            return False
        for child2 in nxdl_elem:
            if get_local_name_from_xml(child) != \
                    get_local_name_from_xml(child2) or get_node_name(child2) == act_htmlname:
                continue
            # check if the name of another sibling fits better
            name_any2 = "nameType" in child2.attrib.keys() and child2.attrib["nameType"] == "any"
            fit2 = get_nx_namefit(chk_name, get_node_name(child2), name_any2)
            if fit2 > fit:
                return False
        # accept this fit
        return True
    return False


def get_local_name_from_xml(element):
    """Helper function to extract the element tag without the namespace."""
    return element.tag[element.tag.rindex("}") + 1:]


def get_own_nxdl_child_reserved_elements(child, name, nxdl_elem):
    """checking reserved elements, like doc, enumeration"""
    if get_local_name_from_xml(child) == 'doc' and name == 'doc':
        if nxdl_elem.get('nxdlbase'):
            child.set('nxdlbase', nxdl_elem.get('nxdlbase'))
            child.set('nxdlbase_class', nxdl_elem.get('nxdlbase_class'))
            child.set('nxdlpath', nxdl_elem.get('nxdlpath') + '/doc')
        return child
    if get_local_name_from_xml(child) == 'enumeration' and name == 'enumeration':
        if nxdl_elem.get('nxdlbase'):
            child.set('nxdlbase', nxdl_elem.get('nxdlbase'))
            child.set('nxdlbase_class', nxdl_elem.get('nxdlbase_class'))
            child.set('nxdlpath', nxdl_elem.get('nxdlpath') + '/enumeration')
        return child
    return False


def get_own_nxdl_child_base_types(child, class_type, nxdl_elem, name, hdf_name):
    """checking base types of group, field,m attribute"""
    if get_local_name_from_xml(child) == 'group':
        if (class_type is None or (class_type and get_nx_class(child) == class_type)) and \
                belongs_to(nxdl_elem, child, name, class_type, hdf_name):
            if nxdl_elem.get('nxdlbase'):
                child.set('nxdlbase', nxdl_elem.get('nxdlbase'))
                child.set('nxdlbase_class', nxdl_elem.get('nxdlbase_class'))
                child.set('nxdlpath', nxdl_elem.get('nxdlpath') + '/' + get_node_name(child))
            return child
    if get_local_name_from_xml(child) == 'field' and \
            belongs_to(nxdl_elem, child, name, None, hdf_name):
        if nxdl_elem.get('nxdlbase'):
            child.set('nxdlbase', nxdl_elem.get('nxdlbase'))
            child.set('nxdlbase_class', nxdl_elem.get('nxdlbase_class'))
            child.set('nxdlpath', nxdl_elem.get('nxdlpath') + '/' + get_node_name(child))
        return child
    if get_local_name_from_xml(child) == 'attribute' and \
            belongs_to(nxdl_elem, child, name, None, hdf_name):
        if nxdl_elem.get('nxdlbase'):
            child.set('nxdlbase', nxdl_elem.get('nxdlbase'))
            child.set('nxdlbase_class', nxdl_elem.get('nxdlbase_class'))
            child.set('nxdlpath', nxdl_elem.get('nxdlpath') + '/' + get_node_name(child))
        return child
    return False


def get_own_nxdl_child(nxdl_elem, name, class_type=None, hdf_name=None, nexus_type=None):
    """Checks if an NXDL child node fits to the specific name (either nxdl or hdf)
        name       - nxdl name
        class_type - nxdl type or hdf classname (for groups, it is obligatory)
        hdf_name   - hdf name"""
    for child in nxdl_elem:
        if 'name' in child.attrib and child.attrib['name'] == name:
            if nxdl_elem.get('nxdlbase'):
                child.set('nxdlbase', nxdl_elem.get('nxdlbase'))
                child.set('nxdlbase_class', nxdl_elem.get('nxdlbase_class'))
                child.set('nxdlpath', nxdl_elem.get('nxdlpath') + '/' + get_node_name(child))
            return child
    for child in nxdl_elem:
        if "name" in child.attrib and child.attrib["name"] == name:
            child.set('nxdlbase', nxdl_elem.get('nxdlbase'))
            return child

    for child in nxdl_elem:
        result = get_own_nxdl_child_reserved_elements(child, name, nxdl_elem)
        if result is not False:
            return result
        if nexus_type and get_local_name_from_xml(child) != nexus_type:
            continue
        result = get_own_nxdl_child_base_types(child, class_type, nxdl_elem, name, hdf_name)
        if result is not False:
            return result
    return None


def find_definition_file(bc_name):
    """find the nxdl file corresponding to the name.
    Note that it first checks in contributed and goes beyond only if no contributed found"""
    bc_filename = None
    for nxdl_folder in ['contributed_definitions', 'base_classes', 'applications']:
        if os.path.exists(f"{get_nexus_definitions_path()}{os.sep}"
                          f"{nxdl_folder}{os.sep}{bc_name}.nxdl.xml"):
            bc_filename = f"{get_nexus_definitions_path()}{os.sep}" \
                          f"{nxdl_folder}{os.sep}{bc_name}.nxdl.xml"
            break
    return bc_filename


def get_nxdl_child(nxdl_elem, name, class_type=None, hdf_name=None, nexus_type=None, go_base=True):  # pylint: disable=too-many-arguments
    """Get the NXDL child node corresponding to a specific name
(e.g. of an HDF5 node,or of a documentation) note that if child is not found in application
definition, it also checks for the base classes"""
    # search for possible fits for hdf_nodes : skipped
    # only exact hits are returned when searching an nxdl child
    own_child = get_own_nxdl_child(nxdl_elem, name, class_type, hdf_name, nexus_type)
    if own_child is not None:
        return own_child
    if not go_base:
        return None
    bc_name = get_nx_class(nxdl_elem)  # check in the base class, app def or contributed
    if bc_name[2] == '_':  # filter primitive types
        return None
    if bc_name == "group":  # Check if it is the root element. Then send to NXroot.nxdl.xml
        bc_name = "NXroot"
    bc_filename = find_definition_file(bc_name)
    if not bc_filename:
        raise ValueError('nxdl file not found in definitions folder!')
    bc_obj = ET.parse(bc_filename).getroot()
    bc_obj.set('nxdlbase', bc_filename)
    if 'category' in bc_obj.attrib:
        bc_obj.set('nxdlbase_class', bc_obj.attrib['category'])
    bc_obj.set('nxdlpath', '')
    return get_own_nxdl_child(bc_obj, name, class_type, hdf_name, nexus_type)


def get_required_string(nxdl_elem):
    """Check for being REQUIRED, RECOMMENDED, OPTIONAL, NOT IN SCHEMA"""
    if nxdl_elem is None:
        return "<<NOT IN SCHEMA>>"
    is_optional = 'optional' in nxdl_elem.attrib.keys() \
        and nxdl_elem.attrib['optional'] == "true"
    is_minoccurs = 'minOccurs' in nxdl_elem.attrib.keys() \
        and nxdl_elem.attrib['minOccurs'] == "0"
    is_recommended = 'recommended' in nxdl_elem.attrib.keys() \
        and nxdl_elem.attrib['recommended'] == "true"

    if is_recommended:
        return "<<RECOMMENDED>>"
    if is_optional or is_minoccurs:
        return "<<OPTIONAL>>"
    # default optionality: in BASE CLASSES is true; in APPLICATIONS is false
    try:
        if nxdl_elem.get('nxdlbase_class') == 'base':
            return "<<OPTIONAL>>"
    except TypeError:
        return "<<REQUIRED>>"
    return "<<REQUIRED>>"


# below there are some functions used in get_nxdl_doc function:
def write_doc_string(logger, doc, attr):
    """Simple function that prints a line in the logger if doc exists"""
    if doc:
        logger.debug("@" + attr + ' [NX_CHAR]')
    return logger, doc, attr


def try_find_units(logger, elem, nxdl_path, doc, attr):
    """Try to find if units is defined inside the field in the NXDL element,
    otherwise try to find if units is defined as a child of the NXDL element."""
    try:  # try to find if units is defined inside the field in the NXDL element
        unit = elem.attrib[attr]
        if doc:
            logger.debug(get_node_concept_path(elem) + "@" + attr + ' [' + unit + ']')
        elem = None
        nxdl_path.append(attr)
    except KeyError:  # otherwise try to find if units is defined as a child of the NXDL element
        orig_elem = elem
        elem = get_nxdl_child(elem, attr, nexus_type='attribute')
        if elem is not None:
            if doc:
                logger.debug(get_node_concept_path(orig_elem)
                             + "@" + attr + ' - [' + get_nx_class(elem) + ']')
            nxdl_path.append(elem)
        else:  # if no units category were defined in NXDL:
            if doc:
                logger.debug(get_node_concept_path(orig_elem)
                             + "@" + attr + " - REQUIRED, but undefined unit category")
            nxdl_path.append(attr)
    return logger, elem, nxdl_path, doc, attr


def check_attr_name_nxdl(param):
    """Check for ATTRIBUTENAME_units in NXDL (normal).
If not defined, check for ATTRIBUTENAME to see if the ATTRIBUTE
is in the SCHEMA, but no units category were defined. """
    (logger, elem, nxdl_path, doc, attr, req_str) = param
    orig_elem = elem
    elem2 = get_nxdl_child(elem, attr, nexus_type='attribute')
    if elem2 is not None:  # check for ATTRIBUTENAME_units in NXDL (normal)
        elem = elem2
        if doc:
            logger.debug(get_node_concept_path(orig_elem)
                         + "@" + attr + ' - [' + get_nx_class(elem) + ']')
        nxdl_path.append(elem)
    else:
        # if not defined, check for ATTRIBUTENAME to see if the ATTRIBUTE
        # is in the SCHEMA, but no units category were defined
        elem2 = get_nxdl_child(elem, attr[:-6], nexus_type='attribute')
        if elem2 is not None:
            req_str = '<<RECOMMENDED>>'
            if doc:
                logger.debug(get_node_concept_path(orig_elem)
                             + "@" + attr + " - RECOMMENDED, but undefined unit category")
            nxdl_path.append(attr)
        else:  # otherwise: NOT IN SCHEMA
            elem = elem2
            if doc:
                logger.debug(get_node_concept_path(orig_elem) + "@" + attr + " - IS NOT IN SCHEMA")
    return logger, elem, nxdl_path, doc, attr, req_str


def try_find_default(logger, orig_elem, elem, nxdl_path, doc, attr):  # pylint: disable=too-many-arguments
    """Try to find if default is defined as a child of the NXDL element """
    if elem is not None:
        if doc:
            logger.debug(get_node_concept_path(orig_elem)
                         + "@" + attr + ' - [' + get_nx_class(elem) + ']')
        nxdl_path.append(elem)
    else:  # if no default category were defined in NXDL:
        if doc:
            logger.debug(get_node_concept_path(orig_elem) + "@" + attr + " - [NX_CHAR]")
        nxdl_path.append(attr)
    return logger, elem, nxdl_path, doc, attr


def other_attrs(logger, orig_elem, elem, nxdl_path, doc, attr):  # pylint: disable=too-many-arguments
    """Handle remaining attributes """
    if elem is not None:
        if doc:
            logger.debug(get_node_concept_path(orig_elem)
                         + "@" + attr + ' - [' + get_nx_class(elem) + ']')
        nxdl_path.append(elem)
    else:
        if doc:
            logger.debug(get_node_concept_path(orig_elem) + "@" + attr + " - IS NOT IN SCHEMA")
    return logger, elem, nxdl_path, doc, attr


def get_node_concept_path(elem):
    """get the short version of nxdlbase:nxdlpath"""
    return str(elem.get('nxdlbase').split('/')[-1] + ":" + elem.get('nxdlpath'))


def get_doc(node, ntype, nxhtml, nxpath):
    """Get documentation"""
    # URL for html documentation
    anchor = ''
    for n_item in nxpath:
        anchor += n_item.lower() + "-"
    anchor = ('https://manual.nexusformat.org/classes/',
              nxhtml + "#" + anchor.replace('_', '-') + ntype)
    if not ntype:
        anchor = anchor[:-1]
    doc = ""  # RST documentation from the field 'doc'
    doc_field = node.find("doc")
    if doc_field is not None:
        doc = doc_field.text
    (index, enums) = get_enums(node)  # enums
    if index:
        enum_str = "\n " + ("Possible values:"
                            if len(enums.split(',')) > 1
                            else "Obligatory value:") + "\n   " + enums + "\n"
    else:
        enum_str = ""
    return anchor, doc + enum_str


def print_doc(node, ntype, level, nxhtml, nxpath):
    """Print documentation"""
    anchor, doc = get_doc(node, ntype, nxhtml, nxpath)
    print("  " * (level + 1) + anchor)
    preferred_width = 80 + level * 2
    wrapper = textwrap.TextWrapper(initial_indent='  ' * (level + 1), width=preferred_width,
                                   subsequent_indent='  ' * (level + 1), expand_tabs=False,
                                   tabsize=0)
    if doc is not None:
        for par in doc.split('\n'):
            print(wrapper.fill(par))


def get_namespace(element):
    """Extracts the namespace for elements in the NXDL"""
    return element.tag[element.tag.index("{"):element.tag.rindex("}") + 1]


def get_enums(node):
    """Makes list of enumerations, if node contains any.
    Returns comma separated STRING of enumeration values, if there are enum tag,
    otherwise empty string."""
    # collect item values from enumeration tag, if any
    namespace = get_namespace(node)
    enums = []
    for enumeration in node.findall(f"{namespace}enumeration"):
        for item in enumeration.findall(f"{namespace}item"):
            enums.append(item.attrib["value"])
        enums = ','.join(enums)
        if enums != "":
            return (True, '[' + enums + ']')
    return (False, "")  # if there is no enumeration tag, returns empty string


def add_base_classes(elist, nx_name=None, elem: ET.Element = None):
    """Add the base classes corresponding to the last eleme in elist to the list. Note that if
elist is empty, a nxdl file with the name of nx_name or a rather room elem is used if provided"""
    if elist and nx_name is None:
        nx_name = get_nx_class(elist[-1])
    # to support recursive defintions, like NXsample in NXsample, the following test is removed
    # if elist and nx_name and f"{nx_name}.nxdl.xml" in (e.get('nxdlbase') for e in elist):
    #     return
    if elem is None:
        if not nx_name:
            return
        nxdl_file_path = find_definition_file(nx_name)
        if nxdl_file_path is None:
            nxdl_file_path = f"{nx_name}.nxdl.xml"
        elem = ET.parse(nxdl_file_path).getroot()
        elem.set('nxdlbase', nxdl_file_path)
    else:
        elem.set('nxdlbase', '')
    if 'category' in elem.attrib:
        elem.set('nxdlbase_class', elem.attrib['category'])
    elem.set('nxdlpath', '')
    elist.append(elem)
    # add inherited base class
    if 'extends' in elem.attrib and elem.attrib['extends'] != 'NXobject':
        add_base_classes(elist, elem.attrib['extends'])
    else:
        add_base_classes(elist)


def set_nxdlpath(child, nxdl_elem):
    """
        Setting up child nxdlbase, nxdlpath and nxdlbase_class from nxdl_element.
    """
    if nxdl_elem.get('nxdlbase'):
        child.set('nxdlbase', nxdl_elem.get('nxdlbase'))
        child.set('nxdlbase_class', nxdl_elem.get('nxdlbase_class'))
        child.set('nxdlpath', nxdl_elem.get('nxdlpath') + '/' + get_node_name(child))
    return child


def get_direct_child(nxdl_elem, html_name):
    """ returns the child of nxdl_elem which has a name
        corresponding to the the html documentation name html_name"""
    for child in nxdl_elem:
        if get_local_name_from_xml(child) in ('group', 'field', 'attribute') and \
                html_name == get_node_name(child):
            decorated_child = set_nxdlpath(child, nxdl_elem)
            return decorated_child
    return None


def get_field_child(nxdl_elem, html_name):
    """ returns the child of nxdl_elem which has a name
        corresponding to the html documentation name html_name"""
    data_child = None
    for child in nxdl_elem:
        if get_local_name_from_xml(child) != 'field':
            continue
        if get_node_name(child) == html_name:
            data_child = set_nxdlpath(child, nxdl_elem)
            break
    return data_child


def get_best_nxdata_child(nxdl_elem, hdf_node, hdf_name):
    """ returns the child of an NXdata nxdl_elem which has a name
        corresponding to the hdf_name"""
    nxdata = hdf_node.parent
    signals = []
    if 'signal' in nxdata.attrs.keys():
        signals.append(nxdata.attrs.get("signal"))
    if "auxiliary_signals" in nxdata.attrs.keys():
        for aux_signal in nxdata.attrs.get("auxiliary_signals"):
            signals.append(aux_signal)
    data_child = get_field_child(nxdl_elem, 'DATA')
    data_error_child = get_field_child(nxdl_elem, 'FIELDNAME_errors')
    for signal in signals:
        if signal == hdf_name:
            return (data_child, 100)
        if hdf_name.endswith('_errors') and signal == hdf_name[:-7]:
            return (data_error_child, 100)
    axes = []
    if "axes" in nxdata.attrs.keys():
        for axis in nxdata.attrs.get("axes"):
            axes.append(axis)
    axis_child = get_field_child(nxdl_elem, 'AXISNAME')
    for axis in axes:
        if axis == hdf_name:
            return (axis_child, 100)
    return (None, 0)


def get_best_child(nxdl_elem, hdf_node, hdf_name, hdf_class_name, nexus_type):
    """ returns the child of nxdl_elem which has a name
        corresponding to the the html documentation name html_name"""
    bestfit = -1
    bestchild = None
    if 'name' in nxdl_elem.attrib.keys() and nxdl_elem.attrib['name'] == 'NXdata' and \
            hdf_node is not None and hdf_node.parent is not None and \
            hdf_node.parent.attrs.get('NX_class') == 'NXdata':
        (fnd_child, fit) = get_best_nxdata_child(nxdl_elem, hdf_node, hdf_name)
        if fnd_child is not None:
            return (fnd_child, fit)
    for child in nxdl_elem:
        fit = -2
        if get_local_name_from_xml(child) == nexus_type and \
                (nexus_type != 'group' or get_nx_class(child) == hdf_class_name):
            name_any = "nameType" in nxdl_elem.attrib.keys() and \
                nxdl_elem.attrib["nameType"] == "any"
            fit = get_nx_namefit(hdf_name, get_node_name(child), name_any)
        if fit > bestfit:
            bestfit = fit
            bestchild = set_nxdlpath(child, nxdl_elem)
    return (bestchild, bestfit)


def walk_elist(elist, html_name):
    """Handle elist from low priority inheritance classes to higher"""
    for ind in range(len(elist) - 1, -1, -1):
        child = get_direct_child(elist[ind], html_name)
        if child is None:
            # check for names fitting to a superclas definition
            main_child = None
            for potential_direct_parent in elist:
                main_child = get_direct_child(potential_direct_parent, html_name)
                if main_child is not None:
                    (fitting_child, _) = get_best_child(elist[ind], None, html_name,
                                                        get_nx_class(main_child),
                                                        get_local_name_from_xml(main_child))
                    if fitting_child is not None:
                        child = fitting_child
                    break
        elist[ind] = child
        if elist[ind] is None:
            del elist[ind]
            continue
        # override: remove low priority inheritance classes if class_type is overriden
        if len(elist) > ind + 1 and get_nx_class(elist[ind]) != get_nx_class(elist[ind + 1]):
            del elist[ind + 1:]
        # add new base class(es) if new element brings such (and not a primitive type)
        if len(elist) == ind + 1 and get_nx_class(elist[ind])[0:3] != 'NX_':
            add_base_classes(elist)
    return elist, html_name


@lru_cache(maxsize=None)
def get_inherited_nodes(nxdl_path: str = None,  # pylint: disable=too-many-arguments,too-many-locals
                        nx_name: str = None, elem: ET.Element = None,
                        attr=False):  # pylint: disable=unused-argument
    """Returns a list of ET.Element for the given path."""
    # let us start with the given definition file
    elist = []  # type: ignore[var-annotated]
    add_base_classes(elist, nx_name, elem)
    nxdl_elem_path = [elist[0]]

    class_path = []  # type: ignore[var-annotated]
    html_path = nxdl_path.split('/')[1:]
    path = html_path
    for pind in range(len(path)):
        html_name = html_path[pind]
        # pylint: disable=duplicate-code
        elist, html_name = walk_elist(elist, html_name)
        if elist:
            class_path.append(get_nx_class(elist[0]))
            nxdl_elem_path.append(elist[0])
    return (class_path, nxdl_elem_path, elist)


def get_node_at_nxdl_path(nxdl_path: str = None,
                          nx_name: str = None, elem: ET.Element = None,
                          exc: bool = True):
    """Returns an ET.Element for the given path.
    This function either takes the name for the NeXus Application Definition
    we are looking for or the root elem from a previously loaded NXDL file
    and finds the corresponding XML element with the needed attributes."""
    try:
        if nxdl_path.count("/") == 1 and nxdl_path not in ("/ENTRY", "/entry"):
            elem = None
            nx_name = "NXroot"
        (class_path, nxdlpath, elist) = get_inherited_nodes(nxdl_path, nx_name, elem)
    except ValueError as value_error:
        if exc:
            raise NxdlAttributeError(f"Attributes were not found for {nxdl_path}. "
                                     "Please check this entry in the template dictionary.") \
                from value_error
        return None
    if class_path and nxdlpath and elist:
        elem = elist[0]
    else:
        elem = None
        if exc:
            raise NxdlAttributeError(f"Attributes were not found for {nxdl_path}. "
                                     "Please check this entry in the template dictionary.")
    return elem



#!/usr/bin/env python

"""
Developers: use this code to develop and test create_release_notes.py
"""
import os

CREDS_FILE = os.path.join(os.environ["HOME"], ".config", "github_token",)

with open(CREDS_FILE, "r") as cf:
    token = cf.read().strip()

from create_release_notes import main

main(base="v2018.5", head="main", milestone="NXDL 2020.1", token=token, debug=True)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Helper functions commonly used by the convert routine."""

from typing import List
from typing import Tuple, Callable, Union
import re
import xml.etree.ElementTree as ET
from datetime import datetime, timezone
import logging

import numpy as np
from ase.data import chemical_symbols
import h5py

from pynxtools import get_nexus_version, get_nexus_version_hash
from pynxtools.nexus import nexus
from pynxtools.nexus.nexus import NxdlAttributeError

logger = logging.getLogger(__name__)


def is_a_lone_group(xml_element) -> bool:
    """Checks whether a given group XML element has no field or attributes mentioned"""
    if xml_element.get("type") == "NXentry":
        return False
    for child in xml_element.findall(".//"):
        if remove_namespace_from_tag(child.tag) in ("field", "attribute"):
            return False
    return True


def get_nxdl_name_from_elem(xml_element) -> str:
    """Extracts the name or uses the type to remove the NX bit and use as name."""
    name_to_add = ""
    if "name" in xml_element.attrib:
        name_to_add = xml_element.attrib["name"]
    elif "type" in xml_element.attrib:
        name_to_add = (f"{convert_nexus_to_caps(xml_element.attrib['type'])}"
                       f"[{convert_nexus_to_suggested_name(xml_element.attrib['type'])}]")
    return name_to_add


def get_all_defined_required_children_for_elem(xml_element):
    """Gets all possible inherited required children for a given NXDL element"""
    list_of_children_to_add = set()
    for child in xml_element:
        child.set("nxdlbase_class", xml_element.get("nxdlbase_class"))
        if child.attrib and get_required_string(child) == "required":
            tag = remove_namespace_from_tag(child.tag)

            name_to_add = get_nxdl_name_from_elem(child)

            if tag in ("field", "attribute"):
                name_to_add = f"@{name_to_add}" if tag == "attribute" else name_to_add
                list_of_children_to_add.add(name_to_add)
                if tag == "field" \
                        and ("units" in child.attrib.keys()
                             and child.attrib["units"] != "NX_UNITLESS"):
                    list_of_children_to_add.add(f"{name_to_add}/@units")
            elif tag == "group":
                nxdlpath = f'{xml_element.get("nxdlpath")}/{get_nxdl_name_from_elem(child)}'
                nxdlbase = xml_element.get("nxdlbase")
                nx_name = nxdlbase[nxdlbase.rfind("/") + 1:nxdlbase.rfind(".nxdl")]
                if nxdlpath not in visited_paths:
                    visited_paths.append(nxdlpath)
                    children = get_all_defined_required_children(nxdlpath, nx_name)
                    further_children = set()
                    for further_child in children:
                        further_children.add(f"{name_to_add}/{further_child}")
                    list_of_children_to_add.update(further_children)
    return list_of_children_to_add


visited_paths: List[str] = []


def get_all_defined_required_children(nxdl_path, nxdl_name):
    """Helper function to find all possible inherited required children for an NXDL path"""
    if nxdl_name == "NXtest":
        return []

    elist = nexus.get_inherited_nodes(nxdl_path, nx_name=nxdl_name)[2]
    list_of_children_to_add = set()
    for elem in elist:
        list_of_children_to_add.update(get_all_defined_required_children_for_elem(elem))

    return list_of_children_to_add


def add_inherited_children(list_of_children_to_add, path, nxdl_root, template):
    """Takes a list of child names and appends them to template for a given path."""
    for child in list_of_children_to_add:
        child_path = f"{path.rsplit('/', 1)[0]}/{child}"
        if child_path not in template.keys():
            optional_parent = check_for_optional_parent(child_path, nxdl_root)
            optionality = "required" if optional_parent == "<<NOT_FOUND>>" else "optional"
            template[optionality][f"{path.rsplit('/', 1)[0]}/{child}"] = None
    return template


def generate_template_from_nxdl(root, template, path="", nxdl_root=None, nxdl_name=None):
    """Helper function to generate a template dictionary for given NXDL"""
    if nxdl_root is None:
        nxdl_name = root.attrib["name"]
        nxdl_root = root
        root = get_first_group(root)

    tag = remove_namespace_from_tag(root.tag)

    if tag == "doc":
        return

    suffix = ""
    if "name" in root.attrib:
        suffix = root.attrib['name']
    elif "type" in root.attrib:
        nexus_class = convert_nexus_to_caps(root.attrib['type'])
        hdf5name = f"[{convert_nexus_to_suggested_name(root.attrib['type'])}]"
        suffix = f"{nexus_class}{hdf5name}"

    path = path + "/" + (f"@{suffix}" if tag == "attribute" else suffix)

    # Only add fields or attributes to the dictionary
    if tag in ("field", "attribute"):
        optionality = get_required_string(root)
        if optionality == "required":
            optional_parent = check_for_optional_parent(path, nxdl_root)
            optionality = "required" if optional_parent == "<<NOT_FOUND>>" else "optional"
            if optional_parent != "<<NOT_FOUND>>":
                template.optional_parents.append(optional_parent)
        template[optionality][path] = None

        # Only add units if it is a field and the the units are defined but not set to NX_UNITLESS
        if tag == "field" \
           and ("units" in root.attrib.keys() and root.attrib["units"] != "NX_UNITLESS"):
            template[optionality][f"{path}/@units"] = None

        parent_path = convert_data_converter_dict_to_nxdl_path(path.rsplit("/", 1)[0])
        list_of_children_to_add = get_all_defined_required_children(parent_path, nxdl_name)
        add_inherited_children(list_of_children_to_add, path, nxdl_root, template)

    elif tag == "group" and is_a_lone_group(root):
        template[get_required_string(root)][path] = None
        template["lone_groups"].append(path)
        path_nxdl = convert_data_converter_dict_to_nxdl_path(path)
        list_of_children_to_add = get_all_defined_required_children(path_nxdl, nxdl_name)
        add_inherited_children(list_of_children_to_add, path, nxdl_root, template)
    # Handling link: link has a target attibute that store absolute path of concept to be
    # linked. Writer reads link from template in the format {'link': <ABSOLUTE PATH>}
    # {'link': ':/<ABSOLUTE PATH TO EXTERNAL FILE>'}
    elif tag == "link":
        # NOTE:  The code below can be implemented later once, NeXus brings optionality in
        # link. Otherwise link will be considered optional by default.

        # optionality = get_required_string(root)
        # optional_parent = check_for_optional_parent(path, nxdl_root)
        # optionality = "required" if optional_parent == "<<NOT_FOUND>>" else "optional"
        # if optionality == "optional":
        #     template.optional_parents.append(optional_parent)
        optionality = "optional"
        template[optionality][path] = {'link': root.attrib['target']}

    for child in root:
        generate_template_from_nxdl(child, template, path, nxdl_root, nxdl_name)


def get_required_string(elem):
    """Helper function to return nicely formatted names for optionality."""
    return nexus.get_required_string(elem)[2:-2].lower()


def convert_nexus_to_caps(nexus_name):
    """Helper function to convert a NeXus class from <NxClass> to <CLASS>."""
    return nexus_name[2:].upper()


def convert_nexus_to_suggested_name(nexus_name):
    """Helper function to suggest a name for a group from its NeXus class."""
    return nexus_name[2:]


def convert_data_converter_entry_to_nxdl_path_entry(entry) -> Union[str, None]:
    """
    Helper function to convert data converter style entry to NXDL style entry:
    ENTRY[entry] -> ENTRY
    """
    regex = re.compile(r'(.*?)(?=\[)')
    results = regex.search(entry)
    return entry if results is None else results.group(1)


def convert_data_converter_dict_to_nxdl_path(path) -> str:
    """
    Helper function to convert data converter style path to NXDL style path:
    /ENTRY[entry]/sample -> /ENTRY/sample
    """
    nxdl_path = ''
    for entry in path.split('/')[1:]:
        nxdl_path += '/' + convert_data_converter_entry_to_nxdl_path_entry(entry)
    return nxdl_path


def get_name_from_data_dict_entry(entry: str) -> str:
    """Helper function to get entry name from data converter style entry

    ENTRY[entry] -> entry
    """
    regex = re.compile(r'(?<=\[)(.*?)(?=\])')
    results = regex.search(entry)
    if results is None:
        return entry
    if entry[0] == "@":
        return "@" + results.group(1)
    return results.group(1)


def convert_data_dict_path_to_hdf5_path(path) -> str:
    """Helper function to convert data converter style path to HDF5 style path

    /ENTRY[entry]/sample -> /entry/sample
    """
    hdf5path = ''
    for entry in path.split('/')[1:]:
        hdf5path += '/' + get_name_from_data_dict_entry(entry)
    return hdf5path


def is_value_valid_element_of_enum(value, elem) -> Tuple[bool, list]:
    """Checks whether a value has to be specific from the NXDL enumeration and returns options."""
    if elem is not None:
        has_enums, enums = nexus.get_enums(elem)
        if has_enums and (isinstance(value, list) or value not in enums[0:-1] or value == ""):
            return False, enums
    return True, []


NUMPY_FLOAT_TYPES = (np.half, np.float16, np.single, np.double, np.longdouble)
NUMPY_INT_TYPES = (np.short, np.intc, np.int_)
NUMPY_UINT_TYPES = (np.ushort, np.uintc, np.uint)

NEXUS_TO_PYTHON_DATA_TYPES = {
    "ISO8601": (str,),
    "NX_BINARY": (bytes, bytearray, np.byte, np.ubyte, np.ndarray),
    "NX_BOOLEAN": (bool, np.ndarray, np.bool_),
    "NX_CHAR": (str, np.ndarray, np.chararray),
    "NX_DATE_TIME": (str,),
    "NX_FLOAT": (float, np.ndarray, np.floating),
    "NX_INT": (int, np.ndarray, np.signedinteger),
    "NX_UINT": (np.ndarray, np.unsignedinteger),
    "NX_NUMBER": (int, float, np.ndarray, np.signedinteger, np.unsignedinteger, np.floating, dict),
    "NX_POSINT": (int, np.ndarray, np.signedinteger),  # > 0 is checked in is_valid_data_field()
    "NX_COMPLEX": (complex, np.ndarray, np.cdouble, np.csingle),
    "NXDL_TYPE_UNAVAILABLE": (str,)  # Defaults to a string if a type is not provided.
}


def check_all_children_for_callable(objects: list, check: Callable, *args) -> bool:
    """Checks whether all objects in list are validated by given callable."""
    for obj in objects:
        if not check(obj, *args):
            return False

    return True


def is_valid_data_type(value, accepted_types):
    """Checks whether the given value or its children are of an accepted type."""
    if not isinstance(value, list):
        return isinstance(value, accepted_types)

    return check_all_children_for_callable(value, isinstance, accepted_types)


def is_positive_int(value):
    """Checks whether the given value or its children are positive."""
    def is_greater_than(num):
        return num.flat[0] > 0 if isinstance(num, np.ndarray) else num > 0

    if isinstance(value, list):
        return check_all_children_for_callable(value, is_greater_than)

    return value.flat[0] > 0 if isinstance(value, np.ndarray) else value > 0


def convert_str_to_bool_safe(value):
    """Only returns True or False if someone mistakenly adds quotation marks but mean a bool.

       For everything else it returns None.
    """
    if value.lower() == "true":
        return True
    if value.lower() == "false":
        return False
    return None


def is_valid_data_field(value, nxdl_type, path):
    """Checks whether a given value is valid according to what is defined in the NXDL.

        This function will also try to convert typical types, for example int to float,
        and return the successful conversion.

        If it fails to convert, it raises an Exception.

        As a default it just returns the value again.
    """
    accepted_types = NEXUS_TO_PYTHON_DATA_TYPES[nxdl_type]

    if not isinstance(value, dict) and not is_valid_data_type(value, accepted_types):
        try:
            if accepted_types[0] is bool and isinstance(value, str):
                value = convert_str_to_bool_safe(value)
                if value is None:
                    raise ValueError
            return accepted_types[0](value)
        except ValueError as exc:
            raise ValueError(f"The value at {path} should be of Python type: {accepted_types}"
                             f", as defined in the NXDL as {nxdl_type}.") from exc

    if nxdl_type == "NX_POSINT" and not is_positive_int(value):
        raise ValueError(f"The value at {path} should be a positive int.")

    if nxdl_type in ("ISO8601", "NX_DATE_TIME"):
        iso8601 = re.compile(r"^(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2}(?:"
                             r"\.\d*)?)(((?!-00:00)(\+|-)(\d{2}):(\d{2})|Z){1})$")
        results = iso8601.search(value)
        if results is None:
            raise ValueError(f"The date at {path} should be a timezone aware ISO8601 "
                             f"formatted str. For example, 2022-01-22T12:14:12.05018Z"
                             f" or 2022-01-22T12:14:12.05018+00:00.")

    return value


def path_in_data_dict(nxdl_path: str, data: dict) -> Tuple[bool, str]:
    """Checks if there is an accepted variation of path in the dictionary & returns the path."""
    for key in data.keys():
        if nxdl_path == convert_data_converter_dict_to_nxdl_path(key):
            return True, key
    return False, None


def check_for_optional_parent(path: str, nxdl_root: ET.Element) -> str:
    """Finds a parent in the branch that is optional and returns it's path or s<<NOT_FOUND>>."""
    parent_path = path.rsplit("/", 1)[0]

    if parent_path == "":
        return "<<NOT_FOUND>>"

    parent_nxdl_path = convert_data_converter_dict_to_nxdl_path(parent_path)
    elem = nexus.get_node_at_nxdl_path(nxdl_path=parent_nxdl_path, elem=nxdl_root)

    if nexus.get_required_string(elem) in ("<<OPTIONAL>>", "<<RECOMMENDED>>"):
        return parent_path

    return check_for_optional_parent(parent_path, nxdl_root)


def is_node_required(nxdl_key, nxdl_root):
    """Checks whether a node at given nxdl path is required"""
    if nxdl_key[nxdl_key.rindex("/") + 1:] == "@units":
        return False
    if nxdl_key[nxdl_key.rindex("/") + 1] == "@":
        nxdl_key = nxdl_key[0:nxdl_key.rindex("/") + 1] + nxdl_key[nxdl_key.rindex("/") + 2:]
    node = nexus.get_node_at_nxdl_path(nxdl_key, elem=nxdl_root, exc=False)
    return nexus.get_required_string(node) == "<<REQUIRED>>"


def all_required_children_are_set(optional_parent_path, data, nxdl_root):
    """Walks over optional parent's children and makes sure all required ones are set"""
    optional_parent_path = convert_data_converter_dict_to_nxdl_path(optional_parent_path)
    for key in data:
        if key in data["lone_groups"]:
            continue
        nxdl_key = convert_data_converter_dict_to_nxdl_path(key)
        if nxdl_key[0:nxdl_key.rfind("/")] == optional_parent_path \
           and is_node_required(nxdl_key, nxdl_root) \
           and data[key] is None:
            return False

    return True


def is_nxdl_path_a_child(nxdl_path: str, parent: str):
    """Takes an NXDL path for an element and an NXDL parent and confirms it is a child."""
    while nxdl_path.rfind("/") != -1:
        nxdl_path = nxdl_path[0:nxdl_path.rfind("/")]
        if parent == nxdl_path:
            return True
    return False


def check_optionality_based_on_parent_group(
        path,
        nxdl_path,
        nxdl_root,
        data,
        template):
    """Checks whether field is part of an optional parent and then confirms its optionality"""
    for optional_parent in template["optional_parents"]:
        optional_parent_nxdl = convert_data_converter_dict_to_nxdl_path(optional_parent)
        if is_nxdl_path_a_child(nxdl_path, optional_parent_nxdl) \
           and not all_required_children_are_set(optional_parent, data, nxdl_root):
            raise LookupError(f"The data entry, {path}, has an optional parent, "
                              f"{optional_parent}, with required children set. Either"
                              f" provide no children for {optional_parent} or provide"
                              f" all required ones.")


def is_group_part_of_path(path_to_group: str, path_of_entry: str) -> bool:
    """Returns true if a group is contained in a path"""
    tokens_group = path_to_group.split("/")
    tokens_entry = convert_data_converter_dict_to_nxdl_path(path_of_entry).split("/")

    if len(tokens_entry) < len(tokens_group):
        return False

    for tog, toe in zip(tokens_group, tokens_entry):
        if tog != toe:
            return False
    return True


def does_group_exist(path_to_group, data):
    """Returns True if the group or any children are set"""
    path_to_group = convert_data_converter_dict_to_nxdl_path(path_to_group)
    for path in data:
        if is_group_part_of_path(path_to_group, path) and data[path] is not None:
            return True
    return False


def ensure_all_required_fields_exist(template, data, nxdl_root):
    """Checks whether all the required fields are in the returned data object."""
    for path in template["required"]:
        entry_name = get_name_from_data_dict_entry(path[path.rindex('/') + 1:])
        if entry_name == "@units":
            continue
        nxdl_path = convert_data_converter_dict_to_nxdl_path(path)
        is_path_in_data_dict, renamed_path = path_in_data_dict(nxdl_path, data)

        renamed_path = path if renamed_path is None else renamed_path
        if path in template["lone_groups"]:
            opt_parent = check_for_optional_parent(path, nxdl_root)
            if opt_parent != "<<NOT_FOUND>>":
                if does_group_exist(opt_parent, data) and not does_group_exist(renamed_path, data):
                    raise ValueError(f"The required group, {path}, hasn't been supplied"
                                     f" while its optional parent, {path}, is supplied.")
                continue
            if not does_group_exist(renamed_path, data):
                raise ValueError(f"The required group, {path}, hasn't been supplied.")
            continue
        if not is_path_in_data_dict or data[renamed_path] is None:
            raise ValueError(f"The data entry corresponding to {path} is required "
                             f"and hasn't been supplied by the reader.")


def try_undocumented(data, nxdl_root: ET.Element):
    """Tries to move entries used that are from base classes but not in AppDef"""
    for path in list(data.undocumented):
        entry_name = get_name_from_data_dict_entry(path[path.rindex('/') + 1:])

        nxdl_path = convert_data_converter_dict_to_nxdl_path(path)

        if entry_name == "@units":
            continue

        if entry_name[0] == "@" and "@" in nxdl_path:
            index_of_at = nxdl_path.rindex("@")
            nxdl_path = nxdl_path[0:index_of_at] + nxdl_path[index_of_at + 1:]

        try:
            elem = nexus.get_node_at_nxdl_path(nxdl_path=nxdl_path, elem=nxdl_root)
            data[get_required_string(elem)][path] = data.undocumented[path]
            del data.undocumented[path]
            units = f"{path}/@units"
            if units in data.undocumented:
                data[get_required_string(elem)][units] = data.undocumented[units]
                del data.undocumented[units]
        except NxdlAttributeError:
            pass


def validate_data_dict(template, data, nxdl_root: ET.Element):
    """Checks whether all the required paths from the template are returned in data dict."""
    assert nxdl_root is not None, "The NXDL file hasn't been loaded."

    # nxdl_path_set helps to skip validation check on the same type of nxdl signiture
    # This reduces huge amount of runing time
    nxdl_path_to_elm: dict = {}

    # Make sure all required fields exist.
    ensure_all_required_fields_exist(template, data, nxdl_root)
    try_undocumented(data, nxdl_root)

    for path in data.get_documented().keys():
        if data[path] is not None:
            entry_name = get_name_from_data_dict_entry(path[path.rindex('/') + 1:])
            nxdl_path = convert_data_converter_dict_to_nxdl_path(path)

            if entry_name == "@units":
                continue

            if entry_name[0] == "@" and "@" in nxdl_path:
                index_of_at = nxdl_path.rindex("@")
                nxdl_path = nxdl_path[0:index_of_at] + nxdl_path[index_of_at + 1:]

            if nxdl_path in nxdl_path_to_elm:
                elem = nxdl_path_to_elm[nxdl_path]
            else:
                elem = nexus.get_node_at_nxdl_path(nxdl_path=nxdl_path, elem=nxdl_root)
                nxdl_path_to_elm[nxdl_path] = elem

            # Only check for validation in the NXDL if we did find the entry
            # otherwise we just pass it along
            if elem is not None \
               and elem.attrib["name"] == entry_name \
               and remove_namespace_from_tag(elem.tag) in ("field", "attribute"):
                check_optionality_based_on_parent_group(path, nxdl_path, nxdl_root, data, template)

                attrib = elem.attrib
                nxdl_type = attrib["type"] if "type" in attrib.keys() else "NXDL_TYPE_UNAVAILABLE"
                data[path] = is_valid_data_field(data[path], nxdl_type, path)
                is_valid_enum, enums = is_value_valid_element_of_enum(data[path], elem)
                if not is_valid_enum:
                    raise ValueError(f"The value at {path} should be on of the "
                                     f"following strings: {enums}")

    return True


def remove_namespace_from_tag(tag):
    """Helper function to remove the namespace from an XML tag."""

    return tag.split("}")[-1]


def get_first_group(root):
    """Helper function to get the actual first group element from the NXDL."""
    for child in root:
        if remove_namespace_from_tag(child.tag) == "group":
            return child
    return root


def check_for_valid_atom_types(atoms: Union[str, list]):
    """Check for whether atom exists in periodic table. """

    if isinstance(atoms, list):
        for elm in atoms:
            if elm not in chemical_symbols:
                raise ValueError(f"The element {elm} is not found in periodictable, "
                                 f"check for correct element name")
    elif isinstance(atoms, str):
        if atoms not in chemical_symbols:
            raise ValueError(f"The element {atoms} is not found in periodictable, "
                             f"check for correct element name")


def convert_to_hill(atoms_typ):
    """Convert list of atom into to hill."""
    if not isinstance(atoms_typ, list) and isinstance(atoms_typ, set):
        atoms_typ = list(atoms_typ)
    atoms_typ = sorted(atoms_typ)
    atom_list = []
    if 'C' in atoms_typ:
        atom_list.append('C')
    if 'H' in atoms_typ:
        atom_list.append('H')
    if atom_list:
        for char in atom_list:
            atoms_typ.remove(char)
    return atom_list + list(atoms_typ)


def add_default_root_attributes(data, filename):
    """
    Takes a dict/Template and adds NXroot fields/attributes that are inherently available
    """
    def update_and_warn(key: str, value: str):
        if key in data and data[key] != value:
            logger.warning(
                "The NXroot entry '%s' (value: %s) should not be populated by the reader. "
                "This is overwritten by the actually used value '%s'",
                key, data[key], value
            )
        data[key] = value

    update_and_warn("/@NX_class", "NXroot")
    update_and_warn("/@file_name", filename)
    update_and_warn("/@file_time", str(datetime.now(timezone.utc).astimezone()))
    update_and_warn("/@file_update_time", data["/@file_time"])
    update_and_warn(
        "/@NeXus_repository",
        "https://github.com/FAIRmat-NFDI/nexus_definitions/"
        f"blob/{get_nexus_version_hash()}"
    )
    update_and_warn("/@NeXus_version", get_nexus_version())
    update_and_warn("/@HDF5_version", '.'.join(map(str, h5py.h5.get_libversion())))
    update_and_warn("/@h5py_version", h5py.__version__)


def extract_atom_types(formula, mode='hill'):
    """Extract atom types form chemical formula."""
    atom_types: set = set()
    element: str = ""

    for char in formula:
        if char.isalpha():
            if char.isupper() and element == "":
                element = char
            elif char.isupper() and element != "" and element.isupper():
                check_for_valid_atom_types(element)
                atom_types.add(element)
                element = char
            elif char.islower() and element.isupper():
                element = element + char
                check_for_valid_atom_types(element)
                atom_types.add(element)
                element = ""

        else:
            if element.isupper():
                check_for_valid_atom_types(element)
                atom_types.add(element)
            element = ""
    if element.isupper():
        atom_types.add(element)

    atom_types = list(atom_types)
    atom_types = sorted(atom_types)

    if mode == 'hill':
        return convert_to_hill(atom_types)

    return atom_types



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Classes representing groups with NeXus-ish formatted data parsed from hspy."""

# pylint: disable=no-member

from typing import Dict

import numpy as np

import hyperspy.api as hs

from pynxtools.dataconverter.readers.em_spctrscpy.utils.em_nexus_base_classes \
    import NxObject


class HspyRectRoiAdfImage:
    """Representing a stack of annular dark field image(s) with metadata."""

    def __init__(self, hspy_clss):
        self.meta: Dict[str, NxObject] = {}
        self.meta["long_name"] = NxObject()
        self.meta["intensity"] = NxObject()
        self.meta["image_id"] = NxObject()
        self.meta["xpos"] = NxObject()
        self.meta["xpos_long_name"] = NxObject()
        self.meta["ypos"] = NxObject()
        self.meta["ypos_long_name"] = NxObject()
        self.is_valid = True

        self.is_supported(hspy_clss)
        self.parse(hspy_clss)

    def is_supported(self, hspy_s2d):
        """Check if the input has supported axes_manager and key metadata."""
        assert hspy_s2d.metadata["Signal"]["signal_type"] == "", \
            "hspy_s2d is not a valid hyperspy generic instance !"
        assert hspy_s2d.data.ndim == 2, \
            "hspy_s2d is not a valid 2D dataset !"
        axes_dict = hspy_s2d.axes_manager.as_dictionary()
        required_axis_names = ["axis-0", "axis-1"]
        for req_key in required_axis_names:
            assert req_key in axes_dict.keys(), \
                req_key + " is unexpectedly not registered in the axes_manager !"
        required_keywords = ["_type", "name", "units", "size", "scale", "offset"]
        avail_axis_names = []
        for keyword in axes_dict.keys():
            for req_key in required_keywords:  # check if all required keys exist
                assert req_key in axes_dict[keyword].keys(), \
                    "hspy_s2d axis " + keyword + " lacks " + req_key + " !"
            assert axes_dict[keyword]["_type"] == "UniformDataAxis", \
                keyword + ", this axis is not of type UniformDataAxis !"
            avail_axis_names.append(axes_dict[keyword]["name"])

        axes_as_expected_emd \
            = np.all(np.sort(avail_axis_names) == np.sort(["y", "x"]))
        axes_as_expected_bcf \
            = np.all(np.sort(avail_axis_names) == np.sort(["height", "width"]))
        # ##MK::Adrien/Cecile"s BCF and EMD example contains at least one
        # such case where the hyperspy created view in metadata is not
        # consistent across representations generated with different parsers
        # this demands adaptive strategies like the one above
        # in the example e.g. the Bruker HAADF image stores dimensions
        # as height and width, while digital micrograph and Velox EMD store
        # y and x... both names are useless without a coordinate system
        # so here discussions with vendors, hspy developers and community are
        # needed!
        if (axes_as_expected_emd is False) and (axes_as_expected_bcf is False):
            print(f"\tIn function {__name__} as expected")
            self.is_valid = False

    def parse(self, hspy_s2d):
        """Parse a hyperspy Signal2D instance into an NX default plottable."""
        # self.long_name.value = hspy_s2d.metadata["Signal"]["signal_type"]
        if self.is_valid is False:
            pass
        print(f"\tIn function {__name__}")
        self.meta["long_name"].value = hspy_s2d.metadata["General"]["title"]
        self.meta["intensity"].value = hspy_s2d.data  # hspy uses numpy and adapts ??
        axes_dict = hspy_s2d.axes_manager.as_dictionary()
        for keyword in axes_dict.keys():
            offset = np.float64(axes_dict[keyword]["offset"])
            scale = np.float64(axes_dict[keyword]["scale"])
            size = np.uint32(axes_dict[keyword]["size"])
            unit = str(axes_dict[keyword]["units"])
            y_axis = (axes_dict[keyword]["name"] == "y") \
                or (axes_dict[keyword]["name"] == "height")
            x_axis = (axes_dict[keyword]["name"] == "x") \
                or (axes_dict[keyword]["name"] == "width")
            if y_axis is True:
                assert axes_dict[keyword]["_type"] == "UniformDataAxis", \
                    keyword + ", this x axis is not of type UniformDataAxis !"
                self.meta["ypos"].value = np.asarray(
                    np.linspace(0., np.float64(size) * scale, num=size,
                                endpoint=True) + offset / 2., np.float64)
                self.meta["ypos"].unit = unit
                self.meta["ypos_long_name"].value = "y"  # ##MK::name y always!
            if x_axis is True:
                assert axes_dict[keyword]["_type"] == "UniformDataAxis", \
                    keyword + ", this y axis is not of type UniformDataAxis !"
                self.meta["xpos"].value = np.asarray(
                    np.linspace(0., np.float64(size) * scale, num=size,
                                endpoint=True) + offset / 2., np.float64)
                self.meta["xpos"].unit = unit
                self.meta["xpos_long_name"].value = "x"  # ##MK::name x always!
            # ##MK::improve case handling
        self.is_valid = True


class NxImageSetEmAdf:
    """Representing a set of (HA)ADF images with metadata."""

    def __init__(self, hspy_list):
        self.meta: Dict[str, NxObject] = {}
        self.meta["program"] = NxObject()
        self.meta["program_version"] = NxObject(is_attr=True)
        self.meta["adf_inner_half_angle"] = NxObject()
        self.meta["adf_outer_half_angle"] = NxObject()
        # an NXdata object, here represented as an instance of HspyRectRoiAdfImage
        self.data = []
        self.is_valid = True

        self.is_an_implemented_case(hspy_list)
        self.parse_hspy_instances(hspy_list)

    def is_an_implemented_case(self, hspy_list):
        """Check if signal instances in a list is a supported combination."""
        # an Xray analysis which this implementation can support needs to be
        # a rectangular ROI. This ROI should have one HAADF overview image
        # one SpectraStack cnts = f(y, x, energy)
        # one accumulated "sum" spectra cnts = f(energy)
        # an arbitrary number of single element X-ray maps cnts = f(y, x)
        cardinality_adf = 0
        for hspy_clss in hspy_list:
            if isinstance(hspy_clss, hs.signals.Signal2D) is True:
                if hspy_clss.metadata["General"]["title"] == "HAADF":
                    cardinality_adf += 1
        if cardinality_adf != 1:
            self.is_valid = False

    def parse_hspy_instances(self, hspy_list):
        """Extract hspy class instances.

        These have to be mappable on NXem base classes
        if these are understood by NOMAD OASIS.
        """
        if self.is_valid is False:
            pass
        print(f"\tIn function {__name__}")
        for hspy_clss in hspy_list:
            if isinstance(hspy_clss, hs.signals.Signal2D) is True:
                ndim = hspy_clss.data.ndim
                if ndim == 2:
                    if hspy_clss.metadata["General"]["title"] == "HAADF":
                        self.data.append(HspyRectRoiAdfImage(hspy_clss))
                        # if self.data[-1].is_valid is False:
                        #     self.is_valid = False

    def report(self, prefix: str, frame_id: int,
               ifo: dict, template: dict) -> dict:
        """Enter data from the NX-specific representation into the template."""
        if self.is_valid is False:
            print(f"\tIn function {__name__} reporting nothing!")
            return template
        print(f"\tIn function {__name__} reporting...")
        assert (len(self.data) >= 0) and (len(self.data) <= 1), \
            "More than one spectrum stack is currently not supported!"

        if len(self.data) == 1:
            trg = f"{prefix}adf/PROCESS[process1]/"
            template[f"{trg}PROGRAM[program1]/program"] = "hyperspy"
            template[f"{trg}PROGRAM[program1]/program/@version"] \
                = hs.__version__
            template[f"{trg}mode"] = "n/a"
            template[f"{trg}detector_identifier"] = "n/a"
            template[f"{trg}source"] = ifo["source_file_name"]
            template[f"{trg}source/@version"] = ifo["source_file_version"]

            # MISSING_DATA_MSG = "not in hspy metadata case specifically in original_metadata"
            # ##MK::!!!!!
            # ##MK::how to communicate that these data do not exist?
            # template[f"{trg}adf_outer_half_angle"] = np.float64(0.)
            # template[f"{trg}adf_outer_half_angle/@units"] = "rad"
            # template[f"{trg}adf_inner_half_angle"] = np.float64(0.)
            # template[f"{trg}adf_inner_half_angle/@units"] = "rad"

            trg = f"{prefix}adf/stack/"
            template[f"{trg}title"] = "Annular dark field image stack"
            # template[f"{trg}@long_name"] = self.data[0].meta["long_name"].value
            template[f"{trg}@signal"] = "data_counts"
            template[f"{trg}@axes"] = ["axis_image_identifier", "axis_y", "axis_x"]
            template[f"{trg}@AXISNAME[axis_x_indices]"] = np.uint32(2)
            template[f"{trg}@AXISNAME[axis_y_indices]"] = np.uint32(1)
            template[f"{trg}@AXISNAME[axis_image_identifier]"] = np.uint32(0)
            template[f"{trg}DATA[data_counts]"] \
                = {"compress": np.reshape(
                    np.atleast_3d(self.data[0].meta["intensity"].value),
                    (1,
                     np.shape(self.data[0].meta["intensity"].value)[0],
                     np.shape(self.data[0].meta["intensity"].value)[1])),
                    "strength": 1}
            # is the data layout correct?
            # I am pretty sure the last two have to be swopped also!!
            template[f"{trg}DATA[data_counts]/@long_name"] = "Counts (a.u.)"
            template[f"{trg}AXISNAME[axis_x]"] \
                = {"compress": self.data[0].meta["xpos"].value, "strength": 1}
            template[f"{trg}AXISNAME[axis_x]/@units"] = self.data[0].meta["xpos"].unit
            template[f"{trg}AXISNAME[axis_x]/@long_name"] \
                = f"x ({self.data[0].meta['xpos'].unit})"
            template[f"{trg}AXISNAME[axis_y]"] \
                = {"compress": self.data[0].meta["ypos"].value, "strength": 1}
            template[f"{trg}AXISNAME[axis_y]/@units"] = self.data[0].meta["ypos"].unit
            template[f"{trg}AXISNAME[axis_y]/@long_name"] \
                = f"y ({self.data[0].meta['ypos'].unit})"
            template[f"{trg}AXISNAME[axis_image_identifier]"] \
                = np.atleast_1d(np.uint32(frame_id))
            template[f"{trg}AXISNAME[axis_image_identifier]/@long_name"] \
                = "image identifier"

        return template



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Convert refractiveindex.info yaml files to nexus"""
from typing import Tuple, Any, Dict

from pynxtools.dataconverter.readers.json_yml.reader import YamlJsonReader
from pynxtools.dataconverter.readers.rii_database.dispersion_reader import (
    DispersionReader,
)
from pynxtools.dataconverter.readers.utils import parse_json, handle_objects


class RiiReader(YamlJsonReader):
    """
    Converts refractiveindex.info yaml files to nexus
    """

    supported_nxdls = ["NXdispersive_material"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.extensions = {
            ".yml": self.read_dispersion,
            ".yaml": self.read_dispersion,
            ".json": self.parse_json_w_fileinfo,
            "default": lambda _: self.appdef_defaults(),
            "objects": self.handle_rii_objects,
        }

    def read_dispersion(self, filename: str):
        """Reads the dispersion from the give filename"""
        download_bibtex = self.kwargs.get('download_bibtex', False)
        return DispersionReader(download_bibtex).read_dispersion(filename)

    def appdef_defaults(self) -> Dict[str, Any]:
        """Fills default entries which are comment for the application
        definition. Like appdef version and name"""
        entries: Dict[str, Any] = {}

        entries["/ENTRY[entry]/definition/@version"] = "0.0.1"
        entries["/ENTRY[entry]/definition/@url"] = ""
        entries["/ENTRY[entry]/definition"] = "NXdispersive_material"

        entries["/@default"] = "entry"
        entries["/ENTRY[entry]/@default"] = "dispersion_x"

        return entries

    def fill_dispersion_in(self, template: Dict[str, Any]):
        """
        Replaces dispersion_x and dispersion_y keys with filenames as their value
        with the parsed dispersion values.
        """
        keys = [f"dispersion_{axis}" for axis in ["y", "z"]]

        for key in keys:
            if key in template:
                template.update(
                    DispersionReader().read_dispersion(template[key], identifier=key)
                )
                del template[key]

    def parse_json_w_fileinfo(self, filename: str) -> Dict[str, Any]:
        """
        Reads json key/value pairs and additionally replaces dispersion_x and _y keys
        with their parsed dispersion values.
        """
        template = parse_json(filename)
        self.fill_dispersion_in(template)

        return template

    def handle_rii_objects(self, objects: Tuple[Any]) -> Dict[str, Any]:
        """Handle objects and generate template entries from them"""
        template = handle_objects(objects)
        self.fill_dispersion_in(template)

        return template


READER = RiiReader



# pylint: disable=too-many-lines
"""Read files from different format and print it in a standard NeXus format
"""

import os

import sys
import logging
import h5py
import click

from pynxtools.nexus.nxdl_utils import *  # pylint: disable=wildcard-import, unused-wildcard-import


def get_nxdl_entry(hdf_info):
    """Get the nxdl application definition for an HDF5 node"""
    entry = hdf_info
    while isinstance(entry['hdf_node'], h5py.Dataset) or \
            'NX_class' not in entry['hdf_node'].attrs.keys() or \
            entry['hdf_node'].attrs['NX_class'] != 'NXentry':
        entry = get_hdf_info_parent(entry)
        if entry['hdf_node'].name == '/':
            return 'NO NXentry found'
    try:
        nxdef = entry['hdf_node']['definition'][()]
        return nxdef.decode()
    except KeyError:  # 'NO Definition referenced'
        return "NXentry"


def get_nx_class_path(hdf_info):
    """Get the full path of an HDF5 node using nexus classes
in case of a field, end with the field name"""
    hdf_node = hdf_info['hdf_node']
    if hdf_node.name == '/':
        return ''
    if isinstance(hdf_node, h5py.Group):
        return get_nx_class_path(get_hdf_info_parent(hdf_info)) + '/' + \
            (hdf_node.attrs['NX_class'] if 'NX_class' in hdf_node.attrs.keys() else
             hdf_node.name.split('/')[-1])
    if isinstance(hdf_node, h5py.Dataset):
        return get_nx_class_path(
            get_hdf_info_parent(hdf_info)) + '/' + hdf_node.name.split('/')[-1]
    return ''


def chk_nxdataaxis_v2(hdf_node, name, logger):
    """Check if dataset is an axis"""
    own_signal = hdf_node.attrs.get('signal')  # check for being a Signal
    if own_signal is str and own_signal == "1":
        logger.debug("Dataset referenced (v2) as NXdata SIGNAL")
    own_axes = hdf_node.attrs.get('axes')  # check for being an axis
    if own_axes is str:
        axes = own_axes.split(':')
        for i in len(axes):
            if axes[i] and name == axes[i]:
                logger.debug("Dataset referenced (v2) as NXdata AXIS #%d", i)
                return None
    ownpaxis = hdf_node.attrs.get('primary')
    own_axis = hdf_node.attrs.get('axis')
    if own_axis is int:
        # also convention v1
        if ownpaxis is int and ownpaxis == 1:
            logger.debug("Dataset referenced (v2) as NXdata AXIS #%d", own_axis - 1)
        else:
            logger.debug(
                "Dataset referenced (v2) as NXdata (primary/alternative) AXIS #%d", own_axis - 1)
    return None


def chk_nxdataaxis(hdf_node, name, logger):
    """NEXUS Data Plotting Standard v3: new version from 2014"""
    if not isinstance(hdf_node, h5py.Dataset):  # check if it is a field in an NXdata node
        return None
    parent = hdf_node.parent
    if not parent or (parent and not parent.attrs.get('NX_class') == "NXdata"):
        return None
    signal = parent.attrs.get('signal')  # chk for Signal
    if signal and name == signal:
        logger.debug("Dataset referenced as NXdata SIGNAL")
        return None
    axes = parent.attrs.get('axes')  # check for default Axes
    if axes is str:
        if name == axes:
            logger.debug("Dataset referenced as NXdata AXIS")
            return None
    elif axes is not None:
        for i, j in enumerate(axes):
            if name == j:
                indices = parent.attrs.get(j + '_indices')
                if indices is int:
                    logger.debug(f"Dataset referenced as NXdata AXIS #{indices}")
                else:
                    logger.debug(f"Dataset referenced as NXdata AXIS #{i}")
                return None
    indices = parent.attrs.get(name + '_indices')  # check for alternative Axes
    if indices is int:
        logger.debug(f"Dataset referenced as NXdata alternative AXIS #{indices}")
    return chk_nxdataaxis_v2(hdf_node, name, logger)  # check for older conventions


def check_deprecation_enum_axis(variables, doc, elist, attr, hdf_node):
    """Check for several attributes. - deprecation - enums - nxdataaxis """
    logger, elem, path = variables
    dep_str = elem.attrib.get('deprecated')  # check for deprecation
    if dep_str:
        if doc:
            logger.debug("DEPRECATED - " + dep_str)
    for base_elem in elist if not attr else [elem]:  # check for enums
        sdoc = get_nxdl_child(base_elem, 'enumeration', go_base=False)
        if sdoc is not None:
            if doc:
                logger.debug("enumeration (" + get_node_concept_path(base_elem) + "):")
            for item in sdoc:
                if get_local_name_from_xml(item) == 'item':
                    if doc:
                        logger.debug("-> " + item.attrib['value'])
    chk_nxdataaxis(hdf_node, path.split('/')[-1], logger)  # look for NXdata reference (axes/signal)
    for base_elem in elist if not attr else [elem]:  # check for doc
        sdoc = get_nxdl_child(base_elem, 'doc', go_base=False)
        if doc:
            logger.debug("documentation (" + get_node_concept_path(base_elem) + "):")
            logger.debug(sdoc.text if sdoc is not None else "")
    return logger, elem, path, doc, elist, attr, hdf_node


def get_nxdl_attr_doc(  # pylint: disable=too-many-arguments,too-many-locals
        elem, elist, attr, hdf_node, logger, doc, nxdl_path, req_str, path, hdf_info):
    """Get nxdl documentation for an attribute"""
    new_elem = []
    old_elem = elem
    for elem_index, act_elem1 in enumerate(elist):
        act_elem = act_elem1
        # NX_class is a compulsory attribute for groups in a nexus file
        # which should match the type of the corresponding NXDL element
        if attr == 'NX_class' and not isinstance(hdf_node, h5py.Dataset) and elem_index == 0:
            elem = None
            logger, doc, attr = write_doc_string(logger, doc, attr)
            new_elem = elem
            break
        # units category is a compulsory attribute for any fields
        if attr == 'units' and isinstance(hdf_node, h5py.Dataset):
            req_str = "<<REQUIRED>>"
            logger, act_elem, nxdl_path, doc, attr = try_find_units(logger,
                                                                    act_elem,
                                                                    nxdl_path,
                                                                    doc,
                                                                    attr)
        # units for attributes can be given as ATTRIBUTENAME_units
        elif attr.endswith('_units'):
            logger, act_elem, nxdl_path, doc, attr, req_str = check_attr_name_nxdl((logger,
                                                                                    act_elem,
                                                                                    nxdl_path,
                                                                                    doc,
                                                                                    attr,
                                                                                    req_str))
        # default is allowed for groups
        elif attr == 'default' and not isinstance(hdf_node, h5py.Dataset):
            req_str = "<<RECOMMENDED>>"
            # try to find if default is defined as a child of the NXDL element
            act_elem = get_nxdl_child(act_elem, attr, nexus_type='attribute', go_base=False)
            logger, act_elem, nxdl_path, doc, attr = try_find_default(logger,
                                                                      act_elem1,
                                                                      act_elem,
                                                                      nxdl_path,
                                                                      doc,
                                                                      attr)
        else:  # other attributes
            act_elem = get_nxdl_child(act_elem, attr, nexus_type='attribute', go_base=False)
            if act_elem is not None:
                logger, act_elem, nxdl_path, doc, attr = \
                    other_attrs(logger, act_elem1, act_elem, nxdl_path, doc, attr)
        if act_elem is not None:
            new_elem.append(act_elem)
            if req_str is None:
                req_str = get_required_string(act_elem)  # check for being required
                if doc:
                    logger.debug(req_str)
            variables = [logger, act_elem, path]
            logger, elem, path, doc, elist, attr, hdf_node = check_deprecation_enum_axis(variables,
                                                                                         doc,
                                                                                         elist,
                                                                                         attr,
                                                                                         hdf_node)
    elem = old_elem
    if req_str is None and doc:
        if attr != 'NX_class':
            logger.debug("@" + attr + " - IS NOT IN SCHEMA")
        logger.debug("")
    return (req_str, get_nxdl_entry(hdf_info), nxdl_path)


def get_nxdl_doc(hdf_info, logger, doc, attr=False):
    """Get nxdl documentation for an HDF5 node (or its attribute)"""
    hdf_node = hdf_info['hdf_node']
    # new way: retrieve multiple inherited base classes
    (class_path, nxdl_path, elist) = \
        get_inherited_hdf_nodes(nx_name=get_nxdl_entry(hdf_info), hdf_node=hdf_node,
                                hdf_path=hdf_info['hdf_path'] if 'hdf_path' in hdf_info else None,
                                hdf_root=hdf_info['hdf_root'] if 'hdf_root' in hdf_info else None)
    elem = elist[0] if class_path and elist else None
    if doc:
        logger.debug("classpath: " + str(class_path))
        logger.debug("NOT IN SCHEMA" if elem is None else
                     "classes:\n" + "\n".join
                     (get_node_concept_path(e) for e in elist))
    # old solution with a single elem instead of using elist
    path = get_nx_class_path(hdf_info)
    req_str = None
    if elem is None:
        if doc:
            logger.debug("")
        return ('None', None, None)
    if attr:
        return get_nxdl_attr_doc(elem, elist, attr, hdf_node, logger, doc, nxdl_path,
                                 req_str, path, hdf_info)
    req_str = get_required_string(elem)  # check for being required
    if doc:
        logger.debug(req_str)
    variables = [logger, elem, path]
    logger, elem, path, doc, elist, attr, hdf_node = check_deprecation_enum_axis(variables,
                                                                                 doc,
                                                                                 elist,
                                                                                 attr,
                                                                                 hdf_node)
    return (req_str, get_nxdl_entry(hdf_info), nxdl_path)


def helper_get_inherited_nodes(hdf_info2, elist, pind, attr):
    """find the best fitting name in all children"""
    hdf_path, hdf_node, hdf_class_path = hdf_info2
    hdf_name = hdf_path[pind]
    hdf_class_name = hdf_class_path[pind]
    if pind < len(hdf_path) - (2 if attr else 1):
        act_nexus_type = 'group'
    elif pind == len(hdf_path) - 1 and attr:
        act_nexus_type = 'attribute'
    else:
        act_nexus_type = 'field' if isinstance(hdf_node, h5py.Dataset) else 'group'
    # find the best fitting name in all children
    bestfit = -1
    html_name = None
    for ind in range(len(elist) - 1, -1, -1):
        newelem, fit = get_best_child(elist[ind],
                                      hdf_node,
                                      hdf_name,
                                      hdf_class_name,
                                      act_nexus_type)
        if fit >= bestfit and newelem is not None:
            html_name = get_node_name(newelem)
    return hdf_path, hdf_node, hdf_class_path, elist, pind, attr, html_name


def get_hdf_path(hdf_info):
    """Get the hdf_path from an hdf_info"""
    if 'hdf_path' in hdf_info:
        return hdf_info['hdf_path'].split('/')[1:]
    return hdf_info['hdf_node'].name.split('/')[1:]


# pylint: disable=too-many-arguments,too-many-locals
@lru_cache(maxsize=None)
def get_inherited_hdf_nodes(nx_name: str = None, elem: ET.Element = None,
                            hdf_node=None, hdf_path=None, hdf_root=None, attr=False):
    """Returns a list of ET.Element for the given path."""
    # let us start with the given definition file
    if hdf_node is None:
        raise ValueError('hdf_node must not be None')
    elist = []  # type: ignore[var-annotated]
    add_base_classes(elist, nx_name, elem)
    nxdl_elem_path = [elist[0]]

    class_path = []  # type: ignore[var-annotated]
    hdf_info = {'hdf_node': hdf_node}
    if hdf_path:
        hdf_info['hdf_path'] = hdf_path
    if hdf_root:
        hdf_root['hdf_root'] = hdf_root
    hdf_node = hdf_info['hdf_node']
    hdf_path = get_hdf_path(hdf_info)
    hdf_class_path = get_nx_class_path(hdf_info).split('/')[1:]
    if attr:
        hdf_path.append(attr)
        hdf_class_path.append(attr)
    path = hdf_path

    for pind in range(len(path)):
        hdf_info2 = [hdf_path, hdf_node, hdf_class_path]
        [hdf_path, hdf_node, hdf_class_path, elist,
            pind, attr, html_name] = helper_get_inherited_nodes(hdf_info2, elist,
                                                                pind, attr)
        if html_name is None:  # return if NOT IN SCHEMA
            return (class_path, nxdl_elem_path, None)
        elist, html_name = walk_elist(elist, html_name)
        if elist:
            class_path.append(get_nx_class(elist[0]))
            nxdl_elem_path.append(elist[0])
    return (class_path, nxdl_elem_path, elist)


def process_node(hdf_node, hdf_path, parser, logger, doc=True):
    """Processes an hdf5 node.
- it logs the node found and also checks for its attributes
- retrieves the corresponding nxdl documentation
TODO:
- follow variants
- NOMAD parser: store in NOMAD """
    hdf_info = {'hdf_path': hdf_path, 'hdf_node': hdf_node}
    if isinstance(hdf_node, h5py.Dataset):
        logger.debug(f'===== FIELD (/{hdf_path}): {hdf_node}')
        val = str(hdf_node[()]).split('\n') if len(hdf_node.shape) <= 1 else str(
            hdf_node[0]).split('\n')
        logger.debug(f'value: {val[0]} {"..." if len(val) > 1 else ""}')
    else:
        logger.debug(
            f"===== GROUP (/{hdf_path} "
            f"[{get_nxdl_entry(hdf_info)}"
            f"::{get_nx_class_path(hdf_info)}]): {hdf_node}"
        )
    (req_str, nxdef, nxdl_path) = get_nxdl_doc(hdf_info, logger, doc)
    if parser is not None and isinstance(hdf_node, h5py.Dataset):
        parser({"hdf_info": hdf_info,
                "nxdef": nxdef,
                "nxdl_path": nxdl_path,
                "val": val,
                "logger": logger})
    for key, value in hdf_node.attrs.items():
        logger.debug(f'===== ATTRS (/{hdf_path}@{key})')
        val = str(value).split('\n')
        logger.debug(f'value: {val[0]} {"..." if len(val) > 1 else ""}')
        (req_str, nxdef, nxdl_path) = \
            get_nxdl_doc(hdf_info, logger, doc, attr=key)
        if (
            parser is not None
            and req_str is not None
            and 'NOT IN SCHEMA' not in req_str
            and 'None' not in req_str
        ):
            parser({"hdf_info": hdf_info,
                    "nxdef": nxdef,
                    "nxdl_path": nxdl_path,
                    "val": val,
                    "logger": logger}, attr=key)


def logger_auxiliary_signal(logger, nxdata):
    """Handle the presence of auxiliary signal"""
    aux = nxdata.attrs.get('auxiliary_signals')
    if aux is not None:
        if isinstance(aux, str):
            aux = [aux]
        for asig in aux:
            logger.debug(f'Further auxiliary signal has been identified: {asig}')
    return logger


def print_default_plotable_header(logger):
    """Print a three-lines header"""
    logger.debug('========================')
    logger.debug('=== Default Plotable ===')
    logger.debug('========================')


def get_default_plotable(root, logger):
    """Get default plotable"""
    print_default_plotable_header(logger)
    # v3 from 2014
    # nxentry
    nxentry = None
    default_nxentry_group_name = root.attrs.get("default")
    if default_nxentry_group_name:
        try:
            nxentry = root[default_nxentry_group_name]
        except KeyError:
            nxentry = None
    if not nxentry:
        nxentry = entry_helper(root)
    if not nxentry:
        logger.debug('No NXentry has been found')
        return
    logger.debug('')
    logger.debug('NXentry has been identified: ' + nxentry.name)
    # nxdata
    nxdata = None
    nxgroup = nxentry
    default_group_name = nxgroup.attrs.get("default")
    while default_group_name:
        try:
            nxgroup = nxgroup[default_group_name]
            default_group_name = nxgroup.attrs.get("default")
        except KeyError:
            pass
    if nxgroup == nxentry:
        nxdata = nxdata_helper(nxentry)
    else:
        nxdata = nxgroup
    if not nxdata:
        logger.debug('No NXdata group has been found')
        return
    logger.debug('')
    logger.debug('NXdata group has been identified: ' + nxdata.name)
    process_node(nxdata, nxdata.name, None, logger, False)
    # signal
    signal = None
    signal_dataset_name = nxdata.attrs.get("signal")
    try:
        signal = nxdata[signal_dataset_name]
    except (TypeError, KeyError):
        signal = None
    if not signal:
        signal = signal_helper(nxdata)
    if not signal:
        logger.debug('No Signal has been found')
        return
    logger.debug('')
    logger.debug('Signal has been identified: ' + signal.name)
    process_node(signal, signal.name, None, logger, False)
    logger = logger_auxiliary_signal(logger, nxdata)  # check auxiliary_signals
    dim = len(signal.shape)
    axes = []  # axes
    axis_helper(dim, nxdata, signal, axes, logger)


def entry_helper(root):
    """Check entry related data"""
    nxentries = []
    for key in root.keys():
        if isinstance(root[key], h5py.Group) and root[key].attrs.get('NX_class') and \
                root[key].attrs['NX_class'] == "NXentry":
            nxentries.append(root[key])
    if len(nxentries) >= 1:
        return nxentries[0]
    return None


def nxdata_helper(nxentry):
    """Check if nxentry hdf5 object has a NX_class and, if it contains NXdata,
return its value"""
    lnxdata = []
    for key in nxentry.keys():
        if isinstance(nxentry[key], h5py.Group) and nxentry[key].attrs.get('NX_class') and \
                nxentry[key].attrs['NX_class'] == "NXdata":
            lnxdata.append(nxentry[key])
    if len(lnxdata) >= 1:
        return lnxdata[0]
    return None


def signal_helper(nxdata):
    """Check signal related data"""
    signals = []
    for key in nxdata.keys():
        if isinstance(nxdata[key], h5py.Dataset):
            signals.append(nxdata[key])
    if len(signals) == 1:  # v3: as there was no selection given, only 1 data field shall exists
        return signals[0]
    if len(signals) > 1:  # v2: select the one with an attribute signal="1" attribute
        for sig in signals:
            if sig.attrs.get("signal") and sig.attrs.get("signal") is str and \
                    sig.attrs.get("signal") == "1":
                return sig
    return None


def find_attrib_axis_actual_dim_num(nxdata, a_item, ax_list):
    """Finds axis that have defined dimensions"""
    # find those with attribute axis= actual dimension number
    lax = []
    for key in nxdata.keys():
        if isinstance(nxdata[key], h5py.Dataset):
            try:
                if nxdata[key].attrs['axis'] == a_item + 1:
                    lax.append(nxdata[key])
            except KeyError:
                pass
    if len(lax) == 1:
        ax_list.append(lax[0])
    # if there are more alternatives, prioritise the one with an attribute primary="1"
    elif len(lax) > 1:
        for sax in lax:
            if sax.attrs.get('primary') and sax.attrs.get('primary') == 1:
                ax_list.insert(0, sax)
            else:
                ax_list.append(sax)


def get_single_or_multiple_axes(nxdata, ax_datasets, a_item, ax_list):
    """Gets either single or multiple axes from the NXDL"""
    try:
        if isinstance(ax_datasets, str):  # single axis is defined
            # explicite definition of dimension number
            ind = nxdata.attrs.get(ax_datasets + '_indices')
            if ind and ind is int:
                if ind == a_item:
                    ax_list.append(nxdata[ax_datasets])
            elif a_item == 0:  # positional determination of the dimension number
                ax_list.append(nxdata[ax_datasets])
        else:  # multiple axes are listed
            # explicite definition of dimension number
            for aax in ax_datasets:
                ind = nxdata.attrs.get(aax + '_indices')
                if ind and isinstance(ind, int):
                    if ind == a_item:
                        ax_list.append(nxdata[aax])
            if not ax_list:  # positional determination of the dimension number
                ax_list.append(nxdata[ax_datasets[a_item]])
    except KeyError:
        pass
    return ax_list


def axis_helper(dim, nxdata, signal, axes, logger):
    """Check axis related data"""
    for a_item in range(dim):
        ax_list = []
        ax_datasets = nxdata.attrs.get("axes")  # primary axes listed in attribute axes
        ax_list = get_single_or_multiple_axes(nxdata, ax_datasets, a_item, ax_list)
        for attr in nxdata.attrs.keys():  # check for corresponding AXISNAME_indices
            if attr.endswith('_indices') and nxdata.attrs[attr] == a_item and \
                    nxdata[attr.split('_indices')[0]] not in ax_list:
                ax_list.append(nxdata[attr.split('_indices')[0]])
        # v2  # check for ':' separated axes defined in Signal
        if not ax_list:
            try:
                ax_datasets = signal.attrs.get("axes").split(':')
                ax_list.append(nxdata[ax_datasets[a_item]])
            except (KeyError, AttributeError):
                pass
        if not ax_list:  # check for axis/primary specifications
            find_attrib_axis_actual_dim_num(nxdata, a_item, ax_list)
        axes.append(ax_list)
        logger.debug('')
        logger.debug(
            f'For Axis #{a_item}, {len(ax_list)} axes have been identified: {str(ax_list)}'
        )


def get_all_is_a_rel_from_hdf_node(hdf_node, hdf_path):
    """Return list of nxdl concept paths for a nxdl element which corresponds to
    hdf node.
    """
    hdf_info = {'hdf_path': hdf_path, 'hdf_node': hdf_node}
    (_, _, elist) = \
        get_inherited_hdf_nodes(nx_name=get_nxdl_entry(hdf_info), hdf_node=hdf_node,
                                hdf_path=hdf_info['hdf_path'] if 'hdf_path' in hdf_info else None,
                                hdf_root=hdf_info['hdf_root'] if 'hdf_root' in hdf_info else None)
    return elist


def hdf_node_to_self_concept_path(hdf_info, logger):
    """ Get concept or nxdl path from given hdf_node.
    """
    # The bellow logger is for deactivatine unnecessary debug message above
    if logger is None:
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.INFO)
    (_, _, nxdl_path) = get_nxdl_doc(hdf_info, logger, None)
    con_path = ''
    if nxdl_path:
        for nd_ in nxdl_path:
            con_path = con_path + '/' + get_node_name(nd_)
    return con_path


class HandleNexus:
    """documentation"""
    def __init__(self, logger, nexus_file,
                 d_inq_nd=None, c_inq_nd=None):
        self.logger = logger
        local_dir = os.path.abspath(os.path.dirname(__file__))

        self.input_file_name = nexus_file if nexus_file is not None else \
            os.path.join(local_dir, '../../tests/data/nexus/201805_WSe2_arpes.nxs')
        self.parser = None
        self.in_file = None
        self.d_inq_nd = d_inq_nd
        self.c_inq_nd = c_inq_nd
        # Aggregating hdf path corresponds to concept query node
        self.hdf_path_list_for_c_inq_nd = []

    def visit_node(self, hdf_name, hdf_node):
        """Function called by h5py that iterates on each node of hdf5file.
        It allows h5py visititems function to visit nodes."""
        if self.d_inq_nd is None and self.c_inq_nd is None:
            process_node(hdf_node, '/' + hdf_name, self.parser, self.logger)
        elif (self.d_inq_nd is not None
              and hdf_name in (self.d_inq_nd, self.d_inq_nd[1:])):
            process_node(hdf_node, '/' + hdf_name, self.parser, self.logger)
        elif self.c_inq_nd is not None:
            attributed_concept = self.c_inq_nd.split('@')
            attr = attributed_concept[1] if len(attributed_concept) > 1 else None
            elist = get_all_is_a_rel_from_hdf_node(hdf_node, '/' + hdf_name)
            if elist is None:
                return
            fnd_superclass = False
            fnd_superclass_attr = False
            for elem in reversed(elist):
                tmp_path = elem.get('nxdlbase').split('.nxdl')[0]
                con_path = '/NX' + tmp_path.split('NX')[-1] + elem.get('nxdlpath')
                if fnd_superclass or con_path == attributed_concept[0]:
                    fnd_superclass = True
                    if attr is None:
                        self.hdf_path_list_for_c_inq_nd.append(hdf_name)
                        break
                    for attribute in hdf_node.attrs.keys():
                        attr_concept = get_nxdl_child(elem, attribute, nexus_type='attribute',
                                                      go_base=False)
                        if attr_concept is not None and \
                                attr_concept.get('nxdlpath').endswith(attr):
                            fnd_superclass_attr = True
                            con_path = '/NX' + tmp_path.split('NX')[-1] \
                                + attr_concept.get('nxdlpath')
                            self.hdf_path_list_for_c_inq_nd.append(hdf_name + "@" + attribute)
                            break
                if fnd_superclass_attr:
                    break

    def not_yet_visited(self, root, name):
        """checking if a new node has already been visited in its path"""
        path = name.split('/')
        for i in range(1, len(path)):
            act_path = '/'.join(path[:i])
            # print(act_path+' - '+name)
            if root['/' + act_path] == root['/' + name]:
                return False
        return True

    def full_visit(self, root, hdf_node, name, func):
        """visiting recursivly all children, but avoiding endless cycles"""
        # print(name)
        if len(name) > 0:
            func(name, hdf_node)
        if isinstance(hdf_node, h5py.Group):
            for ch_name, child in hdf_node.items():
                full_name = ch_name if len(name) == 0 else name + '/' + ch_name
                if self.not_yet_visited(root, full_name):
                    self.full_visit(root, child, full_name, func)

    def process_nexus_master_file(self, parser):
        """Process a nexus master file by processing all its nodes and their attributes"""
        self.parser = parser
        self.in_file = h5py.File(
            self.input_file_name[0]
            if isinstance(self.input_file_name, list)
            else self.input_file_name, 'r'
        )
        self.full_visit(self.in_file, self.in_file, '', self.visit_node)
        if self.d_inq_nd is None and self.c_inq_nd is None:
            get_default_plotable(self.in_file, self.logger)
        # To log the provided concept and concepts founded
        if self.c_inq_nd is not None:
            for hdf_path in self.hdf_path_list_for_c_inq_nd:
                self.logger.info(hdf_path)
        self.in_file.close()


@click.command()
@click.option(
    '-f',
    '--nexus-file',
    required=False,
    default=None,
    help=('NeXus file with extension .nxs to learn NeXus different concept'
          ' documentation and concept.')
)
@click.option(
    '-d',
    '--documentation',
    required=False,
    default=None,
    help=("Definition path in nexus output (.nxs) file. Returns debug"
          "log relavent with that definition path. Example: /entry/data/delays")
)
@click.option(
    '-c',
    '--concept',
    required=False,
    default=None,
    help=("Concept path from application definition file (.nxdl,xml). Finds out"
          "all the available concept definition (IS-A realation) for rendered"
          "concept path. Example: /NXarpes/ENTRY/INSTRUMENT/analyser")
)
def main(nexus_file, documentation, concept):
    """The main function to call when used as a script."""
    logging_format = "%(levelname)s: %(message)s"
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(logging.DEBUG)
    logging.basicConfig(level=logging.INFO, format=logging_format, handlers=[stdout_handler])
    logger = logging.getLogger(__name__)
    logger.addHandler(stdout_handler)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False
    if documentation and concept:
        raise ValueError("Only one option either documentation (-d) or is_a relation "
                         "with a concept (-c) can be requested.")
    nexus_helper = HandleNexus(logger, nexus_file,
                               d_inq_nd=documentation,
                               c_inq_nd=concept)
    nexus_helper.process_nexus_master_file(None)


if __name__ == '__main__':
    main()  # pylint: disable=no-value-for-parameter



import datetime
import json
import os
from pathlib import Path
from typing import Optional

import lxml
import yaml

from ..globals import directories
from ..globals.nxdl import get_nxdl_version
from ..globals.urls import MANUAL_URL
from ..utils.types import PathLike


class AnchorRegistry:
    """Document the NXDL vocabulary. Usage goes as follows

    .. code:: python

        reg = AnchorRegistry(output_path=...)
        # It loaded the existing registry from file (if provided)

        reg.add(...)
        reg.add(...)
        ...
        anchors = reg.flush_anchor_buffer()

        reg.add(...)
        reg.add(...)
        ...
        anchors = reg.flush_anchor_buffer()

        ...
        reg.write()
        # It saved the in-memory registry to file (if provided)
    """

    def __init__(self, output_path: Optional[PathLike] = None) -> None:
        # For example: output_path = get_build_root() / "manual" / "source" / "_static"
        self._writing_enabled = bool(output_path)
        if output_path:
            base = "nxdl_vocabulary"
            output_path = Path(output_path).absolute()
            output_path.mkdir(parents=True, exist_ok=True)
            self._html_file = output_path / f"{base}.html"
            self._txt_file = output_path / f"{base}.txt"
            self._json_file = output_path / f"{base}.json"
            self._yaml_file = output_path / f"{base}.yml"
        else:
            self._html_file = None
            self._txt_file = None
            self._json_file = None
            self._yaml_file = None
        self._registry = self._load_registry()
        self._anchor_buffer = []
        self._nxdl_file = None

    @property
    def all_anchors(self):
        result = []
        for v in self._registry.values():
            result += list(v.keys())
        return result

    @property
    def nxdl_file(self) -> Optional[Path]:
        return self._nxdl_file

    @nxdl_file.setter
    def nxdl_file(self, value: PathLike) -> None:
        self._nxdl_file = Path(value).absolute()

    @property
    def html_url(self):
        rst_file = directories.get_rst_filename(self.nxdl_file)
        manual_root = directories.manual_build_sphinxroot()
        rel_path = rst_file.relative_to(manual_root)
        rel_html = str(rel_path.with_suffix(".html"))
        rel_html = rel_html.replace(os.sep, "/")
        return f"{MANUAL_URL}/{rel_html}"

    def add(self, anchor):
        """Add anchor to the in-memory registry and to the
        current anchor buffer."""
        if anchor not in self._anchor_buffer:
            self._anchor_buffer.append(anchor)

        key = self._key_from_anchor(anchor)

        if key not in self._registry:
            self._registry[key] = {}

        reg = self._registry[key]
        if anchor not in reg:
            hanchor = self._html_anchor(anchor)
            url = f"{self.html_url}{hanchor}"
            reg[anchor] = dict(
                term=anchor,
                html=hanchor,
                url=url,
            )

    def _key_from_anchor(self, anchor):
        key = anchor.lower().split("/")[-1].split("@")[-1].split("-")[0]
        if "@" in anchor:
            # restore preceding "@" symbol
            key = "@" + key
        return key

    def write(self):
        """Write the in-memory registry to files"""
        if not self._writing_enabled:
            return
        contents = dict(
            _metadata=dict(
                datetime=datetime.datetime.utcnow().isoformat(),
                title="NeXus NXDL vocabulary.",
                subtitle="Anchors for all NeXus fields, groups, "
                "attributes, and links.",
                version=get_nxdl_version(),
            ),
            terms=self._registry,
        )
        self._write_yaml(contents)
        self._write_json(contents)
        self._write_txt()
        self._write_html(contents)

    def flush_anchor_buffer(self) -> list:
        """Flush the anchor buffer"""
        self._anchor_buffer, ret = list(), self._anchor_buffer
        return ret

    def _html_anchor(self, anchor):
        """
        Create (internal hyperlink target for) HTML anchor from reST anchor.

        Example:

        * reST anchor: /NXcanSAS/ENTRY/TRANSMISSION_SPECTRUM@timestamp-attribute
        * HTML anchor: #nxcansas-entry-transmission-spectrum-timestamp-attribute
        """
        html_anchor = (
            anchor.lower()
            .lstrip("/")
            .replace("_", "-")
            .replace("@", "-")
            .replace("/", "-")
        )
        return f"#{html_anchor}"

    def _load_registry(self) -> dict:
        """Load the anchor registry in memory."""
        if not self._yaml_file:
            return {}
        registry = None
        if self._yaml_file.exists():
            contents = yaml.load(open(self._yaml_file, "r").read(), Loader=yaml.Loader)
            if contents is not None:
                registry = contents.get("terms")
        return registry or {}

    def _write_html(self, contents):
        """Write the anchors to an HTML file."""
        if not self._html_file:
            return
        root = lxml.etree.Element("html")
        body = lxml.etree.SubElement(root, "body")
        title = lxml.etree.SubElement(body, "h1")
        subtitle = lxml.etree.SubElement(body, "em")

        title.text = contents["_metadata"]["title"].strip(".")
        subtitle.text = contents["_metadata"]["subtitle"].strip(".")
        vocab_list = lxml.etree.SubElement(body, "h2")
        vocab_list.text = "NXDL Vocabulary"

        p = lxml.etree.SubElement(body, "p")
        p.text = "This content is also available in these formats: "
        for ext in "json txt yml".split():
            a = lxml.etree.SubElement(p, "a")
            a.attrib["href"] = f"{MANUAL_URL}/_static/{self._txt_file.stem}.{ext}"
            a.text = f" {ext}"

        dl = lxml.etree.SubElement(body, "dl")
        for term, termlist in sorted(contents["terms"].items()):
            dterm = lxml.etree.SubElement(dl, "dt")
            dterm.text = term
            for _, itemdict in sorted(termlist.items()):
                ddef = lxml.etree.SubElement(dterm, "dd")
                a = lxml.etree.SubElement(ddef, "a")
                a.attrib["href"] = itemdict["url"]
                a.text = itemdict["term"]

        lxml.etree.SubElement(body, "hr")

        foot = lxml.etree.SubElement(body, "p")
        foot_em = lxml.etree.SubElement(foot, "em")
        foot_em.text = f"written: {contents['_metadata']['datetime']}"

        html = lxml.etree.tostring(root, pretty_print=True).decode()
        with open(self._html_file, "w") as f:
            f.write(html)
            f.write("\n")

    def _write_json(self, contents):
        if not self._json_file:
            return
        with open(self._json_file, "w") as f:
            json.dump(contents, f, indent=4, sort_keys=True)
            f.write("\n")

    def _write_txt(self):
        """Compendium (dump the list of all known anchors in raw form)."""
        if not self._txt_file:
            return
        terms = self.all_anchors
        with open(self._txt_file, "w") as f:
            f.write("\n".join(sorted(terms)))
            f.write("\n")

    def _write_yaml(self, contents):
        if not self._yaml_file:
            return
        with open(self._yaml_file, "w") as f:
            yaml.dump(contents, f)



from ..docs import AnchorRegistry
from ..docs import NXClassDocGenerator
from ..docs import XSDDocGenerator
from ..docs.nxdl_index import nxdl_indices
from ..docs.xsd_units import generate_xsd_units_doc
from ..globals import directories
from ..nxdl import iter_definitions
from ..nxdl import validate_definition
from ..utils.copy import copy_files
from ..utils.copy import copydir
from ..utils.diff import diff_ascii
from .nxclass_app import diff_nxclass_docs
from .nxclass_app import save_nxclass_docs


def manual_args(parser):
    parser.add_argument(
        "--test",
        action="store_true",
        help="Validate all NeXus class definitions",
    )
    parser.add_argument(
        "--prepare",
        action="store_true",
        help="Create the build files for the NeXus Impatient Guide",
    )
    parser.add_argument(
        "--diff",
        action="store_true",
        help="Print all changes in the generated documentation",
    )


def manual_exec(args):
    # Copy the documentation source files to the build directory
    if args.prepare:
        copydir(directories.manual_source_root(), directories.manual_build_root())

    # XSD and NXDL document generators
    generate_docs = args.prepare or args.diff
    if generate_docs:
        generator = NXClassDocGenerator()
        xsdgenerator = XSDDocGenerator()
        if args.prepare:
            output_path = directories.manual_build_staticroot()
        else:
            output_path = None
        anchor_registry = AnchorRegistry(output_path=output_path)

    # Generate the NeXus class documentation files in the build directory
    for subdir in ("base_classes", "applications", "contributed_definitions"):
        for nxdl_file in iter_definitions(subdir):
            if args.test:
                validate_definition(nxdl_file)
            if generate_docs:
                rst_lines = generator(nxdl_file, anchor_registry=anchor_registry)
            if args.diff:
                diff_nxclass_docs(nxdl_file, rst_lines)
            if args.prepare:
                save_nxclass_docs(nxdl_file, rst_lines)

    # Generate the NXDL XSD documentation in the build directory
    if generate_docs:
        xsd_file = directories.get_xsd_file()
        rst_lines = xsdgenerator(xsd_file)
        nxdl_desc = directories.manual_build_sphinxroot() / "nxdl_desc.rst"
    if args.diff:
        diff_ascii(xsd_file, rst_lines, nxdl_desc)
    if args.prepare:
        print("generate XSD documentation", nxdl_desc)
        with open(nxdl_desc, "w") as fh:
            fh.writelines(rst_lines)

    if generate_docs:
        xsd_file = directories.get_xsd_units_file()
        rst_lines = generate_xsd_units_doc(xsd_file, "anyUnitsAttr", "units")
        nxdl_desc = directories.manual_build_sphinxroot() / "units.table"
    if args.diff:
        diff_ascii(xsd_file, rst_lines, nxdl_desc)
    if args.prepare:
        print("generate XSD documentation (units)", nxdl_desc)
        with open(nxdl_desc, "w") as fh:
            fh.writelines(rst_lines)

    if generate_docs:
        xsd_file = directories.get_xsd_units_file()
        rst_lines = generate_xsd_units_doc(xsd_file, "primitiveType", "data")
        nxdl_desc = directories.manual_build_sphinxroot() / "types.table"
    if args.diff:
        diff_ascii(xsd_file, rst_lines, nxdl_desc)
    if args.prepare:
        print("generate XSD documentation (types)", nxdl_desc)
        with open(nxdl_desc, "w") as fh:
            fh.writelines(rst_lines)

    # Generate the NeXus class documentation index files in the
    # build directory so the files generated above are included in the docs.
    if generate_docs:
        for name, adict in nxdl_indices().items():
            index_file = adict["index_file"]
            rst_lines = adict["rst_lines"]
            if args.diff:
                diff_ascii(name, rst_lines, index_file)
            if args.prepare:
                print("generate NXDL index", index_file)
                if args.prepare:
                    with open(index_file, "w") as fh:
                        fh.writelines(rst_lines)

    # Generate the anchor list in several format
    if args.prepare:
        print("generate anchor list files in", output_path)
        anchor_registry.write()
        copy_files(EXTRA_FILES)


# Path relative to source directory,
# Path relative to build directory,
# Overwrite (boolean)
EXTRA_FILES = [["NXDL_VERSION", "NXDL_VERSION", True], ["LGPL.txt", "LGPL.txt", True]]



"""This is a code that performs several tests on nexus tool

"""
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import logging
import xml.etree.ElementTree as ET

from pynxtools.nexus import nexus


def test_get_nexus_classes_units_attributes():
    """Check the correct parsing of a separate list for:
Nexus classes (base_classes)
Nexus units (memberTypes)
Nexus attribute type (primitiveTypes)
the tested functions can be found in nexus.py file
"""

    # Test 1
    nexus_classes_list = nexus.get_nx_classes()

    assert 'NXbeam' in nexus_classes_list

    # Test 2
    nexus_units_list = nexus.get_nx_units()
    assert 'NX_TEMPERATURE' in nexus_units_list

    # Test 3
    nexus_attribute_list = nexus.get_nx_attribute_type()
    assert 'NX_FLOAT' in nexus_attribute_list


def test_nexus(tmp_path):
    """
    The nexus test function
    """
    local_dir = os.path.abspath(os.path.dirname(__file__))
    example_data = os.path.join(local_dir, '../data/nexus/201805_WSe2_arpes.nxs')
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    handler = logging.\
        FileHandler(os.path.join(tmp_path, 'nexus_test.log'), 'w')
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    nexus_helper = nexus.HandleNexus(logger, example_data, None, None)
    nexus_helper.process_nexus_master_file(None)

    with open(os.path.join(tmp_path, 'nexus_test.log'), 'r', encoding='utf-8') as logfile:
        log = logfile.readlines()
    with open(
        os.path.join(local_dir, '../data/nexus/Ref_nexus_test.log'),
        'r',
        encoding='utf-8'
    ) as reffile:
        ref = reffile.readlines()
    assert log == ref

    # import filecmp
    # # didn't work with filecmp library
    # log = os.path.join(local_dir, '../data/nexus_test_data/nexus_test.log')
    # ref = os.path.join(local_dir, '../data/nexus_test_data/Ref_nexus_test.log')
    # print(filecmp.cmp(log, ref, shallow=False))

    # print('Testing of nexus.py is SUCCESSFUL.')


def test_get_node_at_nxdl_path():
    """Test to verify if we receive the right XML element for a given NXDL path"""
    local_dir = os.path.abspath(os.path.dirname(__file__))
    nxdl_file_path = os.path.join(local_dir, "../data/dataconverter/NXtest.nxdl.xml")
    elem = ET.parse(nxdl_file_path).getroot()
    node = nexus.get_node_at_nxdl_path("/ENTRY/NXODD_name", elem=elem)
    assert node.attrib["type"] == "NXdata"
    assert node.attrib["name"] == "NXODD_name"

    node = nexus.get_node_at_nxdl_path("/ENTRY/NXODD_name/float_value", elem=elem)
    assert node.attrib["type"] == "NX_FLOAT"
    assert node.attrib["name"] == "float_value"

    node = nexus.get_node_at_nxdl_path("/ENTRY/NXODD_name/AXISNAME/long_name", elem=elem)
    assert node.attrib["name"] == "long_name"

    nxdl_file_path = os.path.join(
        local_dir,
        "../../pynxtools/definitions/contributed_definitions/NXem.nxdl.xml"
    )
    elem = ET.parse(nxdl_file_path).getroot()
    node = nexus.get_node_at_nxdl_path(
        "/ENTRY/measurement/EVENT_DATA_EM/USER/affiliation",
        elem=elem)
    assert node.attrib["name"] == "affiliation"

    node = nexus.get_node_at_nxdl_path("/ENTRY/measurement", elem=elem)
    assert node.attrib["type"] == "NXevent_data_em_set"

    node = nexus.get_node_at_nxdl_path(
        "/ENTRY/measurement/EVENT_DATA_EM/SPECTRUM_SET/summary", elem=elem)
    assert node.attrib["type"] == "NXdata"

    node = nexus.get_node_at_nxdl_path(
        "/ENTRY/measurement/EVENT_DATA_EM/SPECTRUM_SET/summary/DATA", elem=elem)
    assert node.attrib["type"] == "NX_NUMBER"

    node = nexus.get_node_at_nxdl_path(
        "/ENTRY/measurement/EVENT_DATA_EM/SPECTRUM_SET/summary/AXISNAME_indices",
        elem=elem)
    assert node.attrib["name"] == "AXISNAME_indices"

    node = nexus.get_node_at_nxdl_path("/ENTRY/COORDINATE_SYSTEM_SET", elem=elem)
    assert node.attrib["type"] == "NXcoordinate_system_set"

    node = nexus.get_node_at_nxdl_path(
        "/ENTRY/COORDINATE_SYSTEM_SET/TRANSFORMATIONS", elem=elem)
    assert node.attrib["type"] == "NXtransformations"

    node = nexus.get_node_at_nxdl_path(
        "/ENTRY/COORDINATE_SYSTEM_SET/TRANSFORMATIONS/AXISNAME", elem=elem)
    assert node.attrib["type"] == "NX_NUMBER"

    node = nexus.get_node_at_nxdl_path(
        "/ENTRY/COORDINATE_SYSTEM_SET/TRANSFORMATIONS/AXISNAME/transformation_type",
        elem=elem)
    assert node.attrib["name"] == "transformation_type"

    nxdl_file_path = os.path.join(
        local_dir,
        "../../pynxtools/definitions/contributed_definitions/NXiv_temp.nxdl.xml"
    )
    elem = ET.parse(nxdl_file_path).getroot()
    node = nexus.get_node_at_nxdl_path(
        "/ENTRY/INSTRUMENT/ENVIRONMENT/voltage_controller",
        elem=elem)
    assert node.attrib["name"] == "voltage_controller"

    node = nexus.get_node_at_nxdl_path(
        "/ENTRY/INSTRUMENT/ENVIRONMENT/voltage_controller/calibration_time",
        elem=elem)
    assert node.attrib["name"] == "calibration_time"


def test_get_inherited_nodes():
    """Test to verify if we receive the right XML element list for a given NXDL path."""
    local_dir = os.path.abspath(os.path.dirname(__file__))
    nxdl_file_path = os.path.join(
        local_dir,
        "../../pynxtools/definitions/contributed_definitions/NXiv_temp.nxdl.xml"
    )
    elem = ET.parse(nxdl_file_path).getroot()
    (_, _, elist) = nexus.get_inherited_nodes(
        nxdl_path="/ENTRY/INSTRUMENT/ENVIRONMENT",
        elem=elem)
    assert len(elist) == 3

    (_, _, elist) = nexus.get_inherited_nodes(
        nxdl_path="/ENTRY/INSTRUMENT/ENVIRONMENT/voltage_controller",
        elem=elem)
    assert len(elist) == 4

    (_, _, elist) = nexus.get_inherited_nodes(
        nxdl_path="/ENTRY/INSTRUMENT/ENVIRONMENT/voltage_controller",
        nx_name="NXiv_temp")
    assert len(elist) == 4


def test_c_option(tmp_path):
    """
    To check -c option from IV_temp.nxs.
    """

    local_path = os.path.dirname(__file__)
    path_to_ref_files = os.path.join(local_path, '../data/nexus/')
    ref_file = path_to_ref_files + 'Ref1_c_option_test.log'
    tmp_file = os.path.join(tmp_path, 'c_option_1_test.log')
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
    handler = logging.FileHandler(tmp_file, 'w')

    with open(ref_file, encoding='utf-8', mode='r') as ref_f:
        ref = ref_f.readlines()

    handler = logging.FileHandler(tmp_file, 'w')
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter('%(levelname)s: %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    nexus_helper = nexus.HandleNexus(logger, None, None, '/NXbeam')
    nexus_helper.process_nexus_master_file(None)

    with open(tmp_file, encoding='utf-8', mode='r') as tmp_f:
        tmp = tmp_f.readlines()

    assert tmp == ref

    logger.removeHandler(handler)
    handler = logging.FileHandler(tmp_file, 'w')
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter('%(levelname)s: %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    nexus_helper = nexus.HandleNexus(logger, None, None, '/NXdetector/data')
    nexus_helper.process_nexus_master_file(None)

    with open(tmp_file, encoding='utf-8', mode='r') as tmp_f:
        tmp = tmp_f.readlines()
    assert tmp[0] == 'INFO: entry/instrument/analyser/data\n'

    logger.removeHandler(handler)
    handler = logging.FileHandler(tmp_file, 'w')
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter('%(levelname)s: %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    nexus_helper = nexus.HandleNexus(logger, None, None, '/NXdata@signal')
    nexus_helper.process_nexus_master_file(None)

    with open(tmp_file, encoding='utf-8', mode='r') as tmp_f:
        tmp = tmp_f.readlines()
    assert tmp[0] == 'INFO: entry/data@signal\n'


def test_d_option(tmp_path):
    """
    To check -d option for default NXarpes test data file.
    """

    tmp_file = os.path.join(tmp_path, 'd_option_1_test.log')
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    handler = logging.FileHandler(tmp_file, 'w')

    handler = logging.FileHandler(tmp_file, 'w')
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(levelname)s: %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    nexus_helper = nexus.HandleNexus(logger, None, '/entry/instrument/analyser/data', None)
    nexus_helper.process_nexus_master_file(None)

    with open(tmp_file, encoding='utf-8', mode='r') as tmp_f:
        tmp = tmp_f.readlines()

    assert tmp[0] == 'DEBUG: ===== FIELD (//entry/instrument/analyser/data): ' + \
        '<HDF5 dataset "data": shape (80, 146, 195), type "<f4">\n'



#!/usr/bin/env python
'''uses h5py to build the verysimple.nx5 data file'''

import h5py

angle = [18.9094, 18.9096, 18.9098, 18.91,  18.9102, 
         18.9104, 18.9106, 18.9108, 18.911, 18.9112, 
         18.9114, 18.9116, 18.9118, 18.912, 18.9122]
diode = [1193, 4474, 53220, 274310, 515430, 827880, 
         1227100, 1434640, 1330280, 1037070, 598720, 
         316460, 56677, 1000, 1000]

with h5py.File('verysimple.nx5', 'w') as f:
    f.attrs['default'] = 'entry'

    nxentry = f.create_group('entry')
    nxentry.attrs["NX_class"] = 'NXentry'
    nxentry.attrs['default'] = 'data'

    nxdata = nxentry.create_group('data')
    nxdata.attrs["NX_class"] = 'NXdata'
    nxdata.attrs['signal'] = 'counts'
    nxdata.attrs['axes'] = 'two_theta'
    nxdata.attrs['two_theta_indices'] = [0,]

    tth = nxdata.create_dataset('two_theta', data=angle)
    tth.attrs['units'] = 'degrees'
    tth.attrs['long_name'] = 'two_theta (degrees)'

    counts = nxdata.create_dataset('counts', data=diode)
    counts.attrs['units'] = 'counts'
    counts.attrs['long_name'] = 'photodiode counts'
